// USB controller -> ATARI ver 6.00 (UNO R4用)
// Copyright
//   メインプログラム、コントローラタイプ(TYPE_PS4, TYPE_MDmini, TYPE_XBOXONE, TYPE_XBOX360,TYPE_PCENGNmini,TYPE_NEOGEOmini,TYPE_MOUSE1,TYPE_MOUSE2,TYPE_PS3,TYPE_PS5,TYPE_CYBERSTICK,TYPE_ZUIKI1) :
//     HUYE 2018-2024
//   コントローラタイプ(TYPE_PSC, TYPE_RAP3, TYPE_RAP4, TYPE_SNES, TYPE_RETROFREAK, TYPE_EGRETII) : 
//     たねけん 2020,2022-2023
//   モード(IF_MODE_XE1AP) Special thanks to イカ
//
#include <EEPROM.h>
#include "joyuniversal.h"
#include <max_LCD.h>

#define MAX_JOYSTICK 4
#define AD_CNV_DATA1 64
#define AD_CNV_DATA2 192

#define IF_MODE_ATARI_AB 0
#define IF_MODE_ATARI_BA 1
#define IF_MODE_TWINPLUS1_AB 2
#define IF_MODE_TWINPLUS1_BA 3
#define IF_MODE_CYBER 4
#define IF_MODE_XE1AP 5
#define IF_MODE_JOYDRV 6
#define IF_MODE_JAMMA_MD3B 7
#define IF_MODE_TWINPLUS2_AB 8
#define IF_MODE_TWINPLUS2_BA 9
#define IF_MODE_CPSF    10
#define IF_MODE_MAGICAL 11
#define IF_MODE_MD3B 12
#define IF_MODE_MD6B 13
#define IF_MODE_JAMMA_CPSF 14
#define IF_MODE_JAMMA_MAGICAL 15

#define MODE_CHG_TIME 3000 // 3s
#define RENSHA_SPEED 31 // 31ms
#define SW_CHANGE_MARGIN 1500 //1.5s
#define MD6B_RESETTIME 4 // 4ms
#define OUTLIMIT_TIME 64 // 64ms
#define JOYDRV_WAIT  64 // 64ms

#define PORT_IRQ0_ISEL  5 // IELSR 5
#define PORT_IRQ0_VEC  21 // VECTOR 21
#define PORT_IRQ0_IRQ   0 // IRQ 0
#define PORT_IRQ1_ISEL  6 // IELSR 6
#define PORT_IRQ1_VEC  22 // VECTOR 22
#define PORT_IRQ1_IRQ   1 // IRQ 1

#ifdef dobogusinclude
#include <spi4teensy3.h>
#endif
#include <SPI.h>

void put_atari_p1_normal();
void put_atari_p2_normal();
void put_atari_p1_md6b_s1();
void put_atari_p1_md6b_s2();
void put_atari_p2_md6b_s1();
void put_atari_p2_md6b_s2();
void put_atari_p1_joydrv();
void put_atari_p1_cyber();
void put_atari_p2_cyber();

byte flg_chg;
byte gpio_sw1,gpio_sw2;
byte modechg_sts[2];
unsigned long mtimes_now, mtimes_pre;
unsigned int usbtask_time;
volatile byte cyber_speed1, cyber_speed2;
int gpio_cnt;

struct {
  byte data_point;
  byte mask_data;
} directinput_cnvdata[MAX_JOYSTICK][26];

volatile byte md6b_cnt[2];
volatile unsigned int md6b_time[2];
unsigned int mode_time[2];
volatile byte output_time[2];
volatile byte cnv_mode[2];
volatile byte atari_data_SELL[2][6];
volatile byte atari_data_SELH[2][6];
volatile byte md6b_data_SELL[2][6];
volatile byte md6b_data_SELH[2][6];

volatile byte cyber_data[12];

volatile byte joydrv_rcv_portno;
volatile byte joydrv_rcv_motor1;
volatile byte joydrv_rcv_motor2;
volatile byte joydrv_phase_cnt;
volatile byte joydrv_connect_cnt;

volatile byte joydrv_snddata[MAX_JOYSTICK][16];
 //  0: X X X X X X SELECT START
 //  1: R4 R3 R2 R1 L4 L3 L2 L1
 //  2: Z' Z Y X D C B A
 //  3: R右 R左 R下 R上 L右 L左 L下 L上
 //  4: L上下アナログ
 //  5: L左右アナログ
 //  6: R上下アナログ
 //  7: R左右アナログ
 //  8: 左アナログボタン1
 //  9: 左アナログボタン2
 // 10: 左アナログボタン3
 // 11: 左アナログボタン4
 // 12: 右アナログボタン1
 // 13: 右アナログボタン2
 // 14: 右アナログボタン3
 // 15: 右アナログボタン4
volatile struct {
  byte motor1;
  byte motor2;
  byte led_r;
  byte led_g;
  byte led_b;
  byte flush_on;
  byte flush_off;
  bool flg_change;
} stick_ctrldata[MAX_JOYSTICK];
byte rensha_time[2][3];

const byte ps_udlr_data[16] PROGMEM = {0b11111110, 0b11110110, 0b11110111, 0b11110101, 0b11111101, 0b11111001, 0b11111011, 0b11111010,
                                       0b11111111, 0b11111111, 0b11111111, 0b11111111, 0b11111111, 0b11111111, 0b11111111, 0b11111111};

const byte psc_udlr_data[16] PROGMEM = {0b11111010, 0b11111110, 0b11110110, 0b11111111, 0b11111011, 0b11111111, 0b11110111, 0b11111111,
                                        0b11111001, 0b11111101, 0b11110101, 0b11111111, 0b11111111, 0b11111111, 0b11111111, 0b11111111};

char lcd_mode_data[16][17] = {"P1[C]:ATARI_AB  ", "P1[C]:ATARI_BA  ", "P1[C]:TWIN+1C_AB", "P1[C]:TWIN+1C_BA",
                              "P1[C]:CYBERSTICK", "P1[C]:XE-1AP    ", "P1[C]:JOYDRV    ", "P1[C]:JAMMA_MD3B",
                              "P1[C]:TWIN+2C_AB", "P1[C]:TWIN+2C_BA", "P1[C]:CPSF      ", "P1[C]:MAGICAL   ",
                              "P1[C]:MD 3B     ", "P1[C]:MD 6B     ", "P1[C]:JAMMA_CPSF", "P1[C]:JAMMA_MGCL"};

void clr_joydata(byte port, byte mode)
{
  byte i;

  if(mode == 0) {
    for(i=0;i<=3;i++) joydrv_snddata[port][i] = 0b11111111;
  }
  for(i=4;i<=7;i++) joydrv_snddata[port][i] = 0b10000000;
  for(i=8;i<=15;i++) joydrv_snddata[port][i] = 0;
}


class JoystickHID : public JOYUniversal {
public:
  JoystickHID(USB *p) :
  JOYUniversal(p) { };

protected: void wrt_BTN_LPLUSU(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][3] &= 0b11111110;
  else joydrv_snddata[port_no][3] |= 0b00000001;
}

protected: void wrt_BTN_LPLUSD(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][3] &= 0b11111101;
  else joydrv_snddata[port_no][3] |= 0b00000010;
}

protected: void wrt_BTN_LPLUSL(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][3] &= 0b11111011;
  else joydrv_snddata[port_no][3] |= 0b00000100;
}

protected: void wrt_BTN_LPLUSR(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][3] &= 0b11110111;
  else joydrv_snddata[port_no][3] |= 0b00001000;
}

protected: void wrt_BTN_A(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][2] &= 0b11111110;
  else joydrv_snddata[port_no][2] |= 0b00000001;
}

protected: void wrt_BTN_B(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][2] &= 0b11111101;
  else joydrv_snddata[port_no][2] |= 0b00000010;
}

protected: void wrt_BTN_C(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][2] &= 0b11111011;
  else joydrv_snddata[port_no][2] |= 0b00000100;
}

protected: void wrt_BTN_X(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][2] &= 0b11101111;
  else joydrv_snddata[port_no][2] |= 0b00010000;
}

protected: void wrt_BTN_Y(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][2] &= 0b11011111;
  else joydrv_snddata[port_no][2] |= 0b00100000;
}

protected: void wrt_BTN_Z(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][2] &= 0b10111111;
  else joydrv_snddata[port_no][2] |= 0b01000000;
}

protected: void wrt_BTN_L1(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][1] &= 0b11111110;
  else joydrv_snddata[port_no][1] |= 0b00000001;
}

protected: void wrt_BTN_L2(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][1] &= 0b11111101;
  else joydrv_snddata[port_no][1] |= 0b00000010;
}

protected: void wrt_BTN_L3(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][1] &= 0b11111011;
  else joydrv_snddata[port_no][1] |= 0b00000100;
}

protected: void wrt_BTN_R1(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][1] &= 0b11101111;
  else joydrv_snddata[port_no][1] |= 0b00010000;
}

protected: void wrt_BTN_R2(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][1] &= 0b11011111;
  else joydrv_snddata[port_no][1] |= 0b00100000;
}

protected: void wrt_BTN_R3(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][1] &= 0b10111111;
  else joydrv_snddata[port_no][1] |= 0b01000000;
}

protected: void wrt_BTN_START(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][0] &= 0b11111110;
  else joydrv_snddata[port_no][0] |= 0b00000001;
}

protected: void wrt_BTN_SELECT(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][0] &= 0b11111101;
  else joydrv_snddata[port_no][0] |= 0b00000010;
}

protected:
  virtual void ParseHIDData(JOYUniversal *hid, bool is_rpt_id, uint8_t len, uint8_t *buf)
  {
    byte i,j,k;
    byte port_no;
    byte d_pointer;
    uint8_t w_buf[64];

    port_no=byte(PORT_NO);

/*
    if(len > 0) {
      Serial.print(VID, HEX);
      Serial.print(F(":"));
      Serial.print(PID, HEX);
      Serial.println(F(":"));
      for(i=0;i<len;i++) Serial.println(buf[i], HEX);
    }
*/
    // コントローラ別処理
    switch(JOYTYPE) {
      case TYPE_MDmini: // MDmini
        joydrv_snddata[port_no][5] = byte(buf[3]); // 左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[4]); // 上下アナログ

        wrt_BTN_A(port_no, buf[5]&0x0040); // Aボタン
        wrt_BTN_B(port_no, buf[5]&0x0020); // Bボタン
        wrt_BTN_C(port_no, buf[6]&0x0002); // Cボタン
        wrt_BTN_X(port_no, buf[5]&0x0080); // Xボタン
        wrt_BTN_Y(port_no, buf[5]&0x0010); // Yボタン
        wrt_BTN_Z(port_no, buf[6]&0x0001); // Zボタン
        wrt_BTN_START(port_no, buf[6]&0x0020); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[6]&0x0010); // SELECT(MODE)ボタン
        break;

      case TYPE_PS3: // PS3 (DUAL SHOCK3)
        wrt_BTN_LPLUSU(port_no, buf[2]&0x0010); // 上
        wrt_BTN_LPLUSD(port_no, buf[2]&0x0040); // 下
        wrt_BTN_LPLUSL(port_no, buf[2]&0x0080); // 左
        wrt_BTN_LPLUSR(port_no, buf[2]&0x0020); // 右
        wrt_BTN_A(port_no, buf[3]&0x0040); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[3]&0x0020); // Bボタン (circle)
        wrt_BTN_R1(port_no, buf[3]&0x0008); // R1ボタン
        wrt_BTN_R2(port_no, buf[3]&0x0002); // R2ボタン
        wrt_BTN_R3(port_no, buf[2]&0x0004); // R3ボタン
        wrt_BTN_X(port_no, buf[3]&0x0080); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[3]&0x0010); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[3]&0x0004); // L1ボタン
        wrt_BTN_L2(port_no, buf[3]&0x0001); // L2ボタン
        wrt_BTN_L3(port_no, buf[2]&0x0002); // L3ボタン
        wrt_BTN_START(port_no, buf[2]&0x0008); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[2]&0x0001); // SELECTボタン

        joydrv_snddata[port_no][5] = byte(buf[6]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[7]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[8]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[9]); // R上下アナログ
        joydrv_snddata[port_no][8] = byte(buf[18]); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[19]); // 右アナログボタン

        if(stick_ctrldata[port_no].flg_change) {
          for(i = 0; i < PS3_REPORT_BUFFER_SIZE; i++)
            w_buf[i] = pgm_read_byte(&ps3_report_buffer[i]);

          if(stick_ctrldata[port_no].motor1 != 0) {
            w_buf[3] = 0xfe;
            w_buf[4] = stick_ctrldata[port_no].motor1;
          }
          else {
            w_buf[3] = w_buf[4] = 0;
          }

          if(stick_ctrldata[port_no].motor2 != 0) {
            w_buf[1] = 0xfe;
            w_buf[2] = stick_ctrldata[port_no].motor2;
          }
          else {
            w_buf[1] = w_buf[2] = 0;
          }
          SndRpt_ps3(PS3_REPORT_BUFFER_SIZE, w_buf);
        }
        break;

      case TYPE_MYFLASHF300_DI: // PS3 (MAYFLASH ARCADE STICK F300(DINPUT/PS3,XINPUT/NS))
        wrt_BTN_LPLUSU(port_no, (buf[4]==0)); // 上
        wrt_BTN_LPLUSD(port_no, (buf[4]==255)); // 下

        wrt_BTN_LPLUSL(port_no, (buf[3]==0)); // 左
        wrt_BTN_LPLUSR(port_no, (buf[3]==255)); // 右

        wrt_BTN_A(port_no, buf[0]&0x0002); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[0]&0x0004); // Bボタン (circle)
        wrt_BTN_X(port_no, buf[0]&0x0001); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[0]&0x0008); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[0]&0x0010); // L1ボタン (LB)
        wrt_BTN_R1(port_no, buf[0]&0x0020); // R1ボタン (RB)
        wrt_BTN_L2(port_no, buf[0]&0x0040); // L2ボタン (LT)
        wrt_BTN_R2(port_no, buf[0]&0x0080); // R2ボタン (RT)
        wrt_BTN_START(port_no, buf[1]&0x0012); // STARTボタン (SHARE,BACK/VIEW)
        wrt_BTN_SELECT(port_no, buf[1]&0x0001); // SELECTボタン (HOME)
        break;

      case TYPE_MYFLASHF300_XB: // PS3 (MAYFLASH ARCADE STICK F300(XBOX))
        wrt_BTN_LPLUSU(port_no, (buf[4]==0)); // 上
        wrt_BTN_LPLUSD(port_no, (buf[4]==255)); // 下

        wrt_BTN_LPLUSL(port_no, (buf[3]==0)); // 左
        wrt_BTN_LPLUSR(port_no, (buf[3]==255)); // 右

        wrt_BTN_A(port_no, buf[0]&0x0001); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[0]&0x0002); // Bボタン (circle)
        wrt_BTN_X(port_no, buf[0]&0x0008); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[0]&0x0010); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[0]&0x0040); // L1ボタン (LB)
        wrt_BTN_R1(port_no, buf[0]&0x0080); // R1ボタン (RB)
        wrt_BTN_L2(port_no, buf[1]&0x0001); // L2ボタン (LT)
        wrt_BTN_R2(port_no, buf[1]&0x0002); // R2ボタン (RT)
        wrt_BTN_START(port_no, buf[1]&0x0018); // STARTボタン (SHARE,BACK/VIEW)
        wrt_BTN_SELECT(port_no, buf[1]&0x0004); // SELECTボタン (HOME)
        break;

      case TYPE_RAP3: // RAP (PS3mode)
        // [[ HORI Real Arcade Pro.V HAYABUSA(PS3 MODE) ------------------------------------------------------
        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[byte(buf[2])&0b00001111]);

        wrt_BTN_A(port_no, buf[0]&0x0002); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[0]&0x0004); // Bボタン (circle)
        wrt_BTN_R1(port_no, buf[0]&0x0020); // R1ボタン
        wrt_BTN_R2(port_no, buf[0]&0x0080); // R2ボタン
        wrt_BTN_R3(port_no, buf[1]&0x0008); // R3ボタン
        wrt_BTN_X(port_no, buf[0]&0x0001); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[0]&0x0008); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[0]&0x0010); // L1ボタン
        wrt_BTN_L2(port_no, buf[0]&0x0040); // L2ボタン
        wrt_BTN_L3(port_no, buf[1]&0x0004); // L3ボタン
        wrt_BTN_START(port_no, buf[1]&0x0002); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[1]&0x0001); // SELECTボタン
        break;

      case TYPE_PSC:
        // [[ Playstation Classic Controller Parser ----------------------------------------------------------
        joydrv_snddata[port_no][3] = pgm_read_byte_near(&psc_udlr_data[(byte(buf[1])>>2)&0b00001111]);

        wrt_BTN_A(port_no, buf[0]&0x0004); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[0]&0x0002); // Bボタン (circle)
        wrt_BTN_R1(port_no, buf[0]&0x0080); // R1ボタン
        wrt_BTN_R2(port_no, buf[0]&0x0020); // R2ボタン
        wrt_BTN_X(port_no, buf[0]&0x0008); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[0]&0x0001); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[0]&0x0040); // L1ボタン
        wrt_BTN_L2(port_no, buf[0]&0x0010); // L2ボタン
        wrt_BTN_START(port_no, buf[1]&0x0002); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[1]&0x0001); // SELECTボタン
        break;

      case TYPE_SNES:
        // [[ iBUFFALO SNES CLASSIC USB GAMEPAD Parser -------------------------------------------------------
        joydrv_snddata[port_no][5] = byte(buf[0]); // 左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[1]); // 上下アナログ

        wrt_BTN_B(port_no, buf[2]&0x0001); // Aボタン
        wrt_BTN_A(port_no, buf[2]&0x0002); // Bボタン
        wrt_BTN_X(port_no, buf[2]&0x0004); // Xボタン
        wrt_BTN_Y(port_no, buf[2]&0x0008); // Yボタン
        wrt_BTN_L1(port_no, buf[2]&0x0010); // L1ボタン
        wrt_BTN_R1(port_no, buf[2]&0x0020); // R1ボタン
        wrt_BTN_SELECT(port_no, buf[2]&0x0040); // SELECTボタン
        wrt_BTN_START(port_no, buf[2]&0x0080); // STARTボタン
        break;

      case TYPE_RETROFREAK:
        // レトロフリーク コントローラアダプター -------------------------------------------------------------
        joydrv_snddata[port_no][5] = byte(buf[1]); // 左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[2]); // 上下アナログ

        switch(byte(buf[0])) {
          case 1: // NESは未確認
          case 4: // PCE2b PCE 6bは未確認
          case 5: // FC
            wrt_BTN_A(port_no, buf[3]&0x0001); // Aボタン
            wrt_BTN_B(port_no, buf[3]&0x0002); // Bボタン
            wrt_BTN_START(port_no, buf[3]&0x0004); // STARTボタン
            wrt_BTN_SELECT(port_no, buf[3]&0x0008); // SELECTボタン
    
            break;
          case 2: // SFC
            wrt_BTN_B(port_no, buf[3]&0x0001); // Bボタン
            wrt_BTN_A(port_no, buf[3]&0x0002); // Aボタン
            wrt_BTN_Y(port_no, buf[3]&0x0004); // Yボタン
            wrt_BTN_X(port_no, buf[3]&0x0008); // Xボタン
            wrt_BTN_L1(port_no, buf[3]&0x0010); // L1ボタン
            wrt_BTN_R1(port_no, buf[3]&0x0020); // R1ボタン
            wrt_BTN_SELECT(port_no, buf[3]&0x0040); // SELECTボタン
            wrt_BTN_START(port_no, buf[3]&0x0080); // STARTボタン
    
            break;
          case 3: // MegaDrive 6bは未確認
            wrt_BTN_A(port_no, buf[3]&0x0001); // Aボタン
            wrt_BTN_B(port_no, buf[3]&0x0002); // Bボタン
            wrt_BTN_C(port_no, buf[3]&0x0004); // Cボタン
            wrt_BTN_X(port_no, buf[3]&0x0008); // Xボタン
            wrt_BTN_Y(port_no, buf[3]&0x0010); // Yボタン
            wrt_BTN_Z(port_no, buf[3]&0x0020); // Zボタン
            wrt_BTN_START(port_no, buf[3]&0x0040); // STARTボタン
            wrt_BTN_SELECT(port_no, buf[3]&0x0080); // MODEボタン
    
            break;
        }
        break;

      case TYPE_XBOXONE: // XBOX ONE
        if(buf[0] != 0x20) break;

        joydrv_snddata[port_no][5] = byte(buf[11]+128); // L左右アナログ
        joydrv_snddata[port_no][4] = byte((buf[13]+128)^0b11111111); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[15]+128); // R左右アナログ
        joydrv_snddata[port_no][6] = byte((buf[17]+128)^0b11111111); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte((((uint16_t)buf[7] << 8) | buf[6])/4); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte((((uint16_t)buf[9] << 8) | buf[8])/4); // 右アナログボタン

        joydrv_snddata[port_no][3] = byte((buf[5]^0b00001111)|0b11110000); // 十字キー

        wrt_BTN_A(port_no, buf[4]&0x0010); // Aボタン
        wrt_BTN_B(port_no, buf[4]&0x0020); // Bボタン
        wrt_BTN_R1(port_no, buf[5]&0x0020); // R1ボタン
        wrt_BTN_R2(port_no, (joydrv_snddata[port_no][12]>127)); // R2ボタン
        wrt_BTN_R3(port_no, buf[5]&0x0080); // R3ボタン
        wrt_BTN_X(port_no, buf[4]&0x0040); // Xボタン
        wrt_BTN_Y(port_no, buf[4]&0x0080); // Yボタン
        wrt_BTN_L1(port_no, buf[5]&0x0010); // L1ボタン
        wrt_BTN_L2(port_no, (joydrv_snddata[port_no][8]>127)); // L2ボタン
        wrt_BTN_L3(port_no, buf[5]&0x0040); // L3ボタン
        wrt_BTN_START(port_no, buf[4]&0x0004); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[4]&0x0008); // SELECTボタン
        
        if(stick_ctrldata[port_no].flg_change) {
          memset(w_buf, 0, sizeof(w_buf));
          w_buf[0]  = w_buf[3]  = 0x09;
          w_buf[1]  = w_buf[4]  = w_buf[11] = 0x00;
          w_buf[2]  = cmdCounter++;
          w_buf[5]  = 0x0F;
          w_buf[6]  = w_buf[8]  = stick_ctrldata[port_no].motor1;
          w_buf[7]  = w_buf[9]  = stick_ctrldata[port_no].motor2;
          w_buf[10] = w_buf[12] = 0xFF;
          SndRpt(13, w_buf);
        }
        break;

      case TYPE_XBOX360: // XBOX 360
        if(buf[0] != 0x00 || buf[1] != 0x14) break;

        joydrv_snddata[port_no][5] = byte(buf[7]+128); // L左右アナログ
        joydrv_snddata[port_no][4] = byte((buf[9]+128)^0b11111111); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[11]+128); // R左右アナログ
        joydrv_snddata[port_no][6] = byte((buf[13]+128)^0b11111111); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte(buf[4]); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[5]); // 右アナログボタン

        joydrv_snddata[port_no][3] = byte((buf[2]^0b00001111)|0b11110000); // 十字キー

        wrt_BTN_A(port_no, buf[3]&0x0010); // Aボタン
        wrt_BTN_B(port_no, buf[3]&0x0020); // Bボタン
        wrt_BTN_R1(port_no, buf[3]&0x0002); // R1ボタン
        wrt_BTN_R2(port_no, (joydrv_snddata[port_no][12]>127)); // R2ボタン
        wrt_BTN_R3(port_no, buf[2]&0x0080); // R3ボタン
        wrt_BTN_X(port_no, buf[3]&0x0040); // Xボタン
        wrt_BTN_Y(port_no, buf[3]&0x0080); // Yボタン
        wrt_BTN_L1(port_no, buf[3]&0x0001); // L1ボタン
        wrt_BTN_L2(port_no, (joydrv_snddata[port_no][8]>127)); // L2ボタン
        wrt_BTN_L3(port_no, buf[2]&0x0040); // L3ボタン
        wrt_BTN_START(port_no, buf[2]&0x0010); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[2]&0x0020); // SELECTボタン
        
        if(stick_ctrldata[port_no].flg_change) {
          memset(w_buf, 0, sizeof(w_buf));
          w_buf[0]  = w_buf[2]  = w_buf[5]  = w_buf[6]  = w_buf[7]  = 0x00;
          w_buf[1]  = 0x08;
          w_buf[3]  = stick_ctrldata[port_no].motor2;
          w_buf[4]  = stick_ctrldata[port_no].motor1;
          SndRpt(8, w_buf);
        }
        break;

      case TYPE_PCENGNmini: // PCエンジンmini
        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[(byte(buf[2]))&0b00001111]);

        joydrv_snddata[port_no][4] = byte(buf[3]); // L上下アナログ
        joydrv_snddata[port_no][5] = byte(buf[4]); // L左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[5]); // R上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[6]); // R左右アナログ

        wrt_BTN_A(port_no, buf[0]&0x0002); // Aボタン
        wrt_BTN_B(port_no, buf[0]&0x0004); // Bボタン
        wrt_BTN_START(port_no, buf[1]&0x0002); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[1]&0x0001); // SELECTボタン
        break;

      case TYPE_NEOGEOmini: // NEOGEOmini
        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[(byte(buf[3]))&0b00001111]);

        joydrv_snddata[port_no][4] = byte(buf[4]); // L上下アナログ
        joydrv_snddata[port_no][5] = byte(buf[5]); // L左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[6]); // R上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[7]); // R左右アナログ

        wrt_BTN_A(port_no, buf[1]&0x0002); // Aボタン
        wrt_BTN_B(port_no, buf[1]&0x0010); // Bボタン
        wrt_BTN_X(port_no, buf[1]&0x0001); // Xボタン
        wrt_BTN_Y(port_no, buf[1]&0x0008); // Yボタン
        wrt_BTN_START(port_no, buf[2]&0x0008); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[2]&0x0004); // SELECTボタン
        break;

      case TYPE_MOUSE4: // マウス(Type4 アナログ1.5バイト)
        i = byte(buf[1]-128);
        j = byte((((buf[2]>>4)&0x0f)|((buf[3]<<4)&0xf0))-128);
        k = byte(128-buf[4]);
        goto COM_MOUSE;
      case TYPE_MOUSE2: // マウス(Type2 アナログ1バイト)
        i = byte(buf[1]-128);
        j = byte(buf[2]-128);
        k = byte(128-buf[3]);
        goto COM_MOUSE;
      case TYPE_MOUSE1: // マウス(Type1 アナログ2バイト)
        i = byte(buf[1]-128);
        j = byte(buf[3]-128);
        k = byte(128-buf[5]);
       COM_MOUSE:
        joydrv_snddata[port_no][5] = i; // L左右アナログ
        joydrv_snddata[port_no][4] = j; // L上下アナログ
        joydrv_snddata[port_no][6] = k; // R上下アナログ(ホイール)

        wrt_BTN_A(port_no, buf[0]&0x0001); // Aボタン(左クリック)
        wrt_BTN_B(port_no, buf[0]&0x0002); // Bボタン(右クリック)
        wrt_BTN_C(port_no, buf[0]&0x0004); // Cボタン(ホイールボタン)
        wrt_BTN_X(port_no, buf[0]&0x0008); // Xボタン(戻るボタン)
        wrt_BTN_Y(port_no, buf[0]&0x0010); // Yボタン(進むボタン)
//        else {
//          joydrv_snddata[port_no][4] = joydrv_snddata[port_no][5] = joydrv_snddata[port_no][6] = joydrv_snddata[port_no][7] = B10000000;
//        }
        break;

      case TYPE_MOUSE3: // マウス(イーグレットⅡミニ)
        if(buf[2] < 32 || buf[2] >=224) buf[2] = (buf[2]<<2);
        else buf[2] = (buf[2] < 128 ? 127 : 128);
        if(buf[3] < 32 || buf[3] >=224) buf[3] = (buf[3]<<2);
        else buf[3] = (buf[3] < 128 ? 127 : 128);

        joydrv_snddata[port_no][5] = byte(buf[2]-128); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[3]-128); // L上下アナログ
        joydrv_snddata[port_no][6] = byte(128-buf[4]); // R上下アナログ(ホイール)

        wrt_BTN_A(port_no, buf[0]&0x0080); // Aボタン(左大ボタン)
        wrt_BTN_B(port_no, buf[1]&0x0001); // Bボタン(右大ボタン)
        wrt_BTN_X(port_no, buf[0]&0x0002); // Xボタン(真ん中左ボタン)
        wrt_BTN_Y(port_no, buf[0]&0x0010); // Yボタン(真ん中中央ボタン)
        wrt_BTN_Z(port_no, buf[0]&0x0004); // Zボタン(真ん中右ボタン)
        break;

      case TYPE_PS5: // PS5
        if(buf[0] == 0x01) d_pointer = 0;
        else if(buf[0] == 0x31) {
          if (len < 3) break;
          d_pointer = 2;
        }
        else break;

        joydrv_snddata[port_no][5] = byte(buf[d_pointer+1]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[d_pointer+2]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[d_pointer+3]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[d_pointer+4]); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte(buf[d_pointer+5]); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[d_pointer+6]); // 右アナログボタン

        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[byte(buf[d_pointer+8])&0b00001111]);

        wrt_BTN_A(port_no, buf[d_pointer+8]&0x0020); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[d_pointer+8]&0x0040); // Bボタン (circle)
        wrt_BTN_R1(port_no, buf[d_pointer+9]&0x0002); // R1ボタン
        wrt_BTN_R2(port_no, buf[d_pointer+9]&0x0008); // R2ボタン
        wrt_BTN_R3(port_no, buf[d_pointer+9]&0x0080); // R3ボタン
        wrt_BTN_X(port_no, buf[d_pointer+8]&0x0010); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[d_pointer+8]&0x0080); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[d_pointer+9]&0x0001); // L1ボタン
        wrt_BTN_L2(port_no, buf[d_pointer+9]&0x0004); // L2ボタン
        wrt_BTN_L3(port_no, buf[d_pointer+9]&0x0040); // L3ボタン
        wrt_BTN_START(port_no, ((buf[d_pointer+9]&0x0020) || (buf[d_pointer+10]&0x0001))); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[d_pointer+9]&0x0010); // SELECTボタン

        if(stick_ctrldata[port_no].flg_change) {
          memset(w_buf, 0, sizeof(w_buf));
          w_buf[0]  = 0x02;
          w_buf[1]  = 0xFF;
          w_buf[2]  = 0xF7;
          w_buf[3]  = stick_ctrldata[port_no].motor1;
          w_buf[4]  = stick_ctrldata[port_no].motor2;
          w_buf[9]  = w_buf[11]  = w_buf[12]  = w_buf[13]  = w_buf[22]  = w_buf[23]  = w_buf[24]  = w_buf[44]  = 0;
          w_buf[39]  = 3;
          w_buf[42]  = 1;
          w_buf[45]  = stick_ctrldata[port_no].led_r;
          w_buf[46]  = stick_ctrldata[port_no].led_g;
          w_buf[47]  = stick_ctrldata[port_no].led_b;
          SndRpt(sizeof(w_buf), w_buf);
        }
        break;

      case TYPE_EGRETII: // EGRET II Mini Pad
        joydrv_snddata[port_no][5] = byte(buf[2]); // 左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[3]); // 上下アナログ

        wrt_BTN_A(port_no, buf[0]&0x0010); // Aボタン
        wrt_BTN_B(port_no, buf[0]&0x0004); // Bボタン
        wrt_BTN_C(port_no, buf[0]&0x0002); // Cボタン
        wrt_BTN_X(port_no, buf[1]&0x0001); // Xボタン
        wrt_BTN_Y(port_no, buf[0]&0x0008); // Yボタン
        wrt_BTN_Z(port_no, buf[0]&0x0001); // Zボタン
        wrt_BTN_START(port_no, buf[0]&0x0040); // STARTボタン
        wrt_BTN_SELECT(port_no, ((buf[0]&0x0080) || (buf[1]&0x0002))); // SELECTボタン,HOMEボタン
        break;

      case TYPE_ELECOM1: // JC-U4013SBK(DirectInput)
        joydrv_snddata[port_no][5] = byte(buf[3]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[4]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[5]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[6]); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte(buf[17]); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[18]); // 右アナログボタン

        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[byte(buf[2])&0b00001111]);

        wrt_BTN_A(port_no, buf[0]&0x0004); // Aボタン
        wrt_BTN_B(port_no, buf[0]&0x0008); // Bボタン
        wrt_BTN_X(port_no, buf[0]&0x0001); // Xボタン
        wrt_BTN_Y(port_no, buf[0]&0x0002); // Yボタン
        wrt_BTN_L1(port_no, buf[0]&0x0010); // L1ボタン
        wrt_BTN_R1(port_no, buf[0]&0x0020); // R1ボタン
        wrt_BTN_L2(port_no, buf[0]&0x0040); // L2ボタン
        wrt_BTN_R2(port_no, buf[0]&0x0080); // R2ボタン
        wrt_BTN_L3(port_no, buf[1]&0x0001); // L3ボタン
        wrt_BTN_R3(port_no, buf[1]&0x0002); // R3ボタン
        wrt_BTN_SELECT(port_no, ((buf[1]&0x0004) || (buf[1]&0x0010))); // BACKボタン,GUIDEボタン
        wrt_BTN_START(port_no, buf[1]&0x0008); // STARTボタン
        break;

      case TYPE_ELECOM2: // JC-U4013SBK(XInput)
        joydrv_snddata[port_no][5] = byte(128+buf[7]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(127-buf[9]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(128+buf[11]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(127-buf[13]); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte(buf[4]); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[5]); // 右アナログボタン

        joydrv_snddata[port_no][3] = (buf[2]^0b00001111)|0b11110000;

        wrt_BTN_A(port_no, buf[3]&0x0010); // Aボタン
        wrt_BTN_B(port_no, buf[3]&0x0020); // Bボタン
        wrt_BTN_X(port_no, buf[3]&0x0040); // Xボタン
        wrt_BTN_Y(port_no, buf[3]&0x0080); // Yボタン
        wrt_BTN_L1(port_no, buf[3]&0x0001); // L1ボタン
        wrt_BTN_R1(port_no, buf[3]&0x0002); // R1ボタン
        wrt_BTN_L2(port_no, buf[4]>0x007F); // L2ボタン
        wrt_BTN_R2(port_no, (buf[5]>0x007f)); // R2ボタン
        wrt_BTN_L3(port_no, buf[2]&0x0040); // L3ボタン
        wrt_BTN_R3(port_no, buf[2]&0x0080); // R3ボタン
        wrt_BTN_SELECT(port_no, ((buf[2]&0x0020) || (buf[3]&0x0004))); // BACKボタン,GUIDEボタン
        wrt_BTN_START(port_no, buf[2]&0x0010); // STARTボタン
        break;

      case TYPE_ELECOM3: // SYSGRATION ELECOM JC-U3312
        joydrv_snddata[port_no][5] = byte(buf[0]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[1]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[2]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[3]); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte(buf[16]); // L1アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[17]); // R1アナログボタン
        joydrv_snddata[port_no][9]  = byte(buf[18]); // L2アナログボタン
        joydrv_snddata[port_no][13] = byte(buf[19]); // R2アナログボタン
        wrt_BTN_LPLUSU(port_no, buf[8]); // POV上
        wrt_BTN_LPLUSD(port_no, buf[10]); // POV下
        wrt_BTN_LPLUSL(port_no, buf[11]); // POV左
        wrt_BTN_LPLUSR(port_no, buf[9]); // POV右
        wrt_BTN_A(port_no, buf[14]); // Aボタン(cross:3)
        wrt_BTN_B(port_no, buf[13]); // Bボタン(circle:4)
        wrt_BTN_X(port_no, buf[15]); // Xボタン(square:1)
        wrt_BTN_Y(port_no, buf[12]); // Yボタン(triangle:2)
        wrt_BTN_L1(port_no, buf[5]&0x0001); // L1ボタン(5)
        wrt_BTN_R1(port_no, buf[5]&0x0002); // R1ボタン(6)
        wrt_BTN_L2(port_no, buf[5]&0x0004); // L2ボタン(7)
        wrt_BTN_R2(port_no, buf[5]&0x0008); // R2ボタン(8)
        wrt_BTN_L3(port_no, buf[5]&0x0040); // L3ボタン(11)
        wrt_BTN_R3(port_no, buf[5]&0x0080); // R3ボタン(12)
        wrt_BTN_SELECT(port_no, buf[5]&0x0010); // SELECTボタン(9)
        wrt_BTN_START(port_no, buf[5]&0x0020); // STARTボタン(10)
        break;

      case TYPE_ELECOM4: // SYSGRATION ELECOM JC-U2912TBK tane
        joydrv_snddata[port_no][5] = byte(buf[2]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[3]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[5]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[4]); // R上下アナログ

        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[byte(buf[1]>>4)&0b00001111]);
        
        wrt_BTN_A(port_no, buf[0]&0x0004); // Aボタン(cross:3)
        wrt_BTN_B(port_no, buf[0]&0x0008); // Bボタン(circle:4)
        wrt_BTN_X(port_no, buf[0]&0x0001); // Xボタン(square:1)
        wrt_BTN_Y(port_no, buf[0]&0x0002); // Yボタン(triangle:2)
        wrt_BTN_L1(port_no, buf[0]&0x0010); // L1ボタン(5)
        wrt_BTN_R1(port_no, buf[0]&0x0020); // R1ボタン(6)
        wrt_BTN_L2(port_no, buf[0]&0x0040); // L2ボタン(7)
        wrt_BTN_R2(port_no, buf[0]&0x0080); // R2ボタン(8)
        wrt_BTN_L3(port_no, buf[1]&0x0001); // L3ボタン(9)
        wrt_BTN_R3(port_no, buf[1]&0x0002); // R3ボタン(10)
        wrt_BTN_SELECT(port_no, buf[1]&0x0004); // SELECTボタン(11)
        wrt_BTN_START(port_no, buf[1]&0x0008); // STARTボタン(12)
        break;

      case TYPE_CYBERSTICK: // CYBER STICK
        joydrv_snddata[port_no][4] = byte(buf[6]^0b11111111); // L上下アナログ
        joydrv_snddata[port_no][5] = byte(buf[5]); // L左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[4]); // R上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[3]); // R左右アナログ

        wrt_BTN_R1(port_no, buf[0]&0x0001); // Aボタン(R1)
        wrt_BTN_R2(port_no, buf[0]&0x0002); // Bボタン(R2)
        wrt_BTN_L1(port_no, buf[0]&0x0004); // Cボタン(L1)
        wrt_BTN_L2(port_no, buf[0]&0x0008); // Dボタン(L2)
        wrt_BTN_X(port_no, buf[0]&0x0010); // E1ボタン(X)
        wrt_BTN_Y(port_no, buf[0]&0x0020); // E2ボタン(Y)
        wrt_BTN_START(port_no, buf[0]&0x0080); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[0]&0x0040); // SELECTボタン
        break;

      case TYPE_SANWA1: // JY-P35U
        joydrv_snddata[port_no][5] = byte(buf[0]); // 左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[1]); // 上下アナログ

        wrt_BTN_A(port_no, buf[2]&0x0008); // Aボタン
        wrt_BTN_B(port_no, buf[2]&0x0010); // Bボタン
        wrt_BTN_C(port_no, buf[2]&0x0020); // Cボタン
        wrt_BTN_X(port_no, buf[2]&0x0001); // Xボタン
        wrt_BTN_Y(port_no, buf[2]&0x0002); // Yボタン
        wrt_BTN_Z(port_no, buf[2]&0x0004); // Zボタン
        wrt_BTN_L1(port_no, buf[2]&0x0040); // L1ボタン
        wrt_BTN_R1(port_no, buf[2]&0x0080); // R1ボタン
        break;

      case TYPE_ZUIKI1: // X68000Z
        joydrv_snddata[port_no][5] = byte(buf[2]); // 左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[3]); // 上下アナログ

        wrt_BTN_A(port_no, buf[0]&0x0002); // Aボタン(表記はb)
        wrt_BTN_B(port_no, buf[0]&0x0004); // Bボタン(表記はa)
        wrt_BTN_L1(port_no, buf[0]&0x0010); // L1ボタン
        wrt_BTN_R1(port_no, buf[0]&0x0020); // R1ボタン
        wrt_BTN_START(port_no, buf[1]&0x0002); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[1]&0x0001); // SELECTボタン
        break;

      case TYPE_8BITDO_TYPE1_DI:  // 8BIT DO Direct Input
        joydrv_snddata[port_no][5] = byte(buf[4]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[5]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[6]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[7]); // R上下アナログ

        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[byte(buf[3])]);

        wrt_BTN_A(port_no, buf[1]&0x0001); // Aボタン
        wrt_BTN_B(port_no, buf[1]&0x0002); // Bボタン
        wrt_BTN_X(port_no, buf[1]&0x0008); // Xボタン
        wrt_BTN_Y(port_no, buf[1]&0x0010); // Yボタン
        wrt_BTN_L1(port_no, buf[1]&0x0040); // L1ボタン
        wrt_BTN_L2(port_no, buf[2]&0x0001); // L2ボタン
        wrt_BTN_L3(port_no, buf[2]&0x0020); // L3ボタン
        wrt_BTN_R1(port_no, buf[1]&0x0080); // R1ボタン
        wrt_BTN_R2(port_no, buf[2]&0x0002); // R2ボタン
        wrt_BTN_R3(port_no, buf[2]&0x0040); // R3ボタン
        wrt_BTN_START(port_no, buf[2]&0x0018); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[2]&0x0004); // SELECTボタン
        break;

      case TYPE_PS4: // PS4
        if(buf[0] == 0x01) d_pointer = 0;
        else if(buf[0] == 0x11) {
          if (len < 4) break;
          d_pointer = 2;
        }
        else break;

        joydrv_snddata[port_no][5] = byte(buf[d_pointer+1]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[d_pointer+2]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[d_pointer+3]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[d_pointer+4]); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte(buf[d_pointer+8]); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[d_pointer+9]); // 右アナログボタン

        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[byte(buf[d_pointer+5])&0b00001111]);

        wrt_BTN_A(port_no, buf[d_pointer+5]&0x0020); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[d_pointer+5]&0x0040); // Bボタン (circle)
        wrt_BTN_R1(port_no, buf[d_pointer+6]&0x0002); // R1ボタン
        wrt_BTN_R2(port_no, buf[d_pointer+6]&0x0008); // R2ボタン
        wrt_BTN_R3(port_no, buf[d_pointer+6]&0x0080); // R3ボタン
        wrt_BTN_X(port_no, buf[d_pointer+5]&0x0010); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[d_pointer+5]&0x0080); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[d_pointer+6]&0x0001); // L1ボタン
        wrt_BTN_L2(port_no, buf[d_pointer+6]&0x0004); // L2ボタン
        wrt_BTN_L3(port_no, buf[d_pointer+6]&0x0040); // L3ボタン
        wrt_BTN_START(port_no, ((buf[d_pointer+6]&0x0020) || (buf[d_pointer+7]&0x0001))); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[d_pointer+6]&0x0010); // SELECTボタン

        if(stick_ctrldata[port_no].flg_change) {
          memset(w_buf, 0, sizeof(w_buf));
          w_buf[0]  = 0x05;
          w_buf[1]  = 0xFF;
          w_buf[4]  = stick_ctrldata[port_no].motor2;
          w_buf[5]  = stick_ctrldata[port_no].motor1;
          w_buf[6]  = stick_ctrldata[port_no].led_r;
          w_buf[7]  = stick_ctrldata[port_no].led_g;
          w_buf[8]  = stick_ctrldata[port_no].led_b;
          w_buf[9]  = stick_ctrldata[port_no].flush_on;
          w_buf[10] = stick_ctrldata[port_no].flush_off;
          SndRpt(sizeof(w_buf), w_buf);
        }
        break;

      case TYPE_DIRECTINPUT: // DirectInput コントローラ

        break;
    }
    stick_ctrldata[port_no].flg_change = false;
  }
};

USB Usb;
//volatile JOYHub Hub(&Usb);
volatile JoystickHID Hid0(&Usb);
volatile JoystickHID Hid1(&Usb);
volatile JoystickHID Hid2(&Usb);
volatile JoystickHID Hid3(&Usb);
volatile JoystickHID Hid4(&Usb);
Max_LCD lcd(&Usb);

void setup() {
  byte i;

//  Serial.begin(115200); // serialプリントを使うときだけ有効にする
//  while (!Serial); // serialプリントを使うときだけ有効にする

  cyber_speed1 = cyber_speed2 = 72;
  for(i=0;i<6;i++) atari_data_SELL[0][i] = atari_data_SELH[0][i] = atari_data_SELL[1][i] = atari_data_SELH[1][i] = md6b_data_SELL[0][i] = md6b_data_SELH[0][i] = md6b_data_SELL[1][i] = md6b_data_SELH[1][i] = 0b00000101;

  for(i=0;i<MAX_JOYSTICK;i++) {
    clr_joydata(i, 0);
    stick_ctrldata[i].motor1=0;
    stick_ctrldata[i].motor2=0;
    stick_ctrldata[i].led_r=0;
    stick_ctrldata[i].led_g=0;
    stick_ctrldata[i].led_b=0;
    stick_ctrldata[i].flush_on=0;
    stick_ctrldata[i].flush_off=0;
    stick_ctrldata[i].flg_change=false;
  }

  /* 時間処理変数の初期化 */
  md6b_time[0] = md6b_time[1] = mode_time[0] = mode_time[1] = 0;
  rensha_time[0][0] = rensha_time[0][1]= rensha_time[0][2] = rensha_time[1][0] = rensha_time[1][1] = rensha_time[1][2] = 254;
  output_time[0] = output_time[1] = 0;
  usbtask_time = 0;

  joydrv_phase_cnt = 0;
  joydrv_connect_cnt = 0;

  delay(10);

  if (Usb.Init() == -1) {
//    Serial.print(F("\r\nOSC did not start"));
    while (1);
  }

  lcd.begin(16,2);
  lcd.noBlink();
  lcd.noCursor();
  lcd.clear();

  modechg_sts[0] = EEPROM.read(2);
  if(modechg_sts[0] > 1) modechg_sts[0] = 0;
  if(modechg_sts[0] == 0) {
    cnv_mode[0] = EEPROM.read(0);
    if(cnv_mode[0] > 15) cnv_mode[0] = IF_MODE_ATARI_AB;
  }
  modechg_sts[1] = EEPROM.read(3);
  if(modechg_sts[1] > 1) modechg_sts[1] = 0;
  if(modechg_sts[1] == 0) {
    cnv_mode[1] = EEPROM.read(1);
    if(cnv_mode[1] > 15  || cnv_mode[1] == IF_MODE_JOYDRV || cnv_mode[1] == IF_MODE_TWINPLUS2_AB || cnv_mode[1] == IF_MODE_JAMMA_CPSF || cnv_mode[1] == IF_MODE_JAMMA_MD3B || cnv_mode[1] == IF_MODE_TWINPLUS2_BA || cnv_mode[1] == IF_MODE_JAMMA_MAGICAL)
      cnv_mode[1] = IF_MODE_ATARI_AB;
  }
  gpio_sw1 = gpio_sw2 = Usb.gpioRd();
  gpio_sw1 &= 0x0f;
  gpio_sw2 = (gpio_sw2>>4)&0x0f;
  gpio_cnt = -1;
  flg_chg = 1;  // モード切替状態として、モードのLED初期表示を行う

//  delay(500);

  mtimes_now = mtimes_pre= millis();
}

byte set_cnv_mode(byte port)
{
  byte i,mode;
  volatile uint32_t *irq_ptr = (volatile uint32_t *)SCB->VTOR;

  volatile byte *irq_pri;
  irq_pri = (volatile byte *)0xE000E400;

  // UNO R4
  R_PFS->PORT[1].PIN[11].PmnPFS = 0x00000004; // P111 = Output
  delay(750);

  mode = ((port == 1 && (cnv_mode[0] == IF_MODE_JOYDRV || cnv_mode[0] == IF_MODE_TWINPLUS2_AB || cnv_mode[0] == IF_MODE_JAMMA_CPSF || cnv_mode[0] == IF_MODE_JAMMA_MD3B || cnv_mode[0] == IF_MODE_TWINPLUS2_BA || cnv_mode[0] == IF_MODE_JAMMA_MAGICAL)) ? cnv_mode[0] : cnv_mode[port]);

  if(modechg_sts[port] == 1) {
    for(i=0;i<5;i++) {
      // UNO R4
      R_PFS->PORT[1].PIN[11].PmnPFS_BY = 0b00000101; // P111 = Output, P111= H, LED13 ON

      delay(50);
      // UNO R4
      R_PFS->PORT[1].PIN[11].PmnPFS_BY = 0b00000100; // P111 = Output, P111= H, LED13 OFF
      delay(50);
    }
    delay(450);
  }

  if(mode&0x01) {
    // UNO R4
    R_PFS->PORT[1].PIN[11].PmnPFS_BY = 0b00000101; // P111 = Output, P111= H, LED13 ON

    delay(1000);

    // UNO R4
    R_PFS->PORT[1].PIN[11].PmnPFS_BY = 0b00000100; // P111 = Output, P111= H, LED13 OFF
    delay(500);
  }
  i = ((mode>>1)&0x07)+1;
  for(;i>0;i--) {
    // UNO R4
    R_PFS->PORT[1].PIN[11].PmnPFS_BY = 0b00000101; // P111 = Output, P111= H, LED13 ON

    delay(250);
    // UNO R4
    R_PFS->PORT[1].PIN[11].PmnPFS_BY = 0b00000100; // P111 = Output, P111= H, LED13 OFF

    delay(250);
  }
  delay(700);

  md6b_cnt[port] = 0;
  switch(mode) {
    case IF_MODE_ATARI_AB:
    case IF_MODE_ATARI_BA:
    case IF_MODE_TWINPLUS1_AB:
    case IF_MODE_TWINPLUS1_BA:
    case IF_MODE_CPSF:
    case IF_MODE_MAGICAL:
    case IF_MODE_MD3B:
      if(port == 0) {
        // UNO R4
        R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000101; // P014 = Output, P014= H, ATARI_P1_1
        R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000101; // P000 = Output, P000= H, ATARI_P1_2
        R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000101; // P001 = Output, P001= H, ATARI_P1_3
        R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000101; // P002 = Output, P002= H, ATARI_P1_4
        R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P1_6
        R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = 0b00000101; // P100 = Output, P100= H, ATARI_P1_7

        attachInterrupt(digitalPinToInterrupt(2), put_atari_p1_normal, CHANGE);
        *(irq_ptr+PORT_IRQ0_VEC) = (uint32_t)put_atari_p1_normal;
        R_ICU->IRQCR[PORT_IRQ0_IRQ] = 0x02; // デジタルフィルタ無効, PCLKB, 両エッジ
      }
      else {
        // UNO R4
        R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = 0b00000101; // P103 = Output, P103= H, ATARI_P2_1
        R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = 0b00000101; // P102 = Output, P102= H, ATARI_P2_2
        R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = 0b00000101; // P106 = Output, P106= H, ATARI_P2_3
        R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = 0b00000101; // P107 = Output, P107= H, ATARI_P2_4
        R_PFS->PORT[3].PIN[ 4].PmnPFS_BY = 0b00000101; // P304 = Output, P304= H, ATARI_P2_6
        R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = 0b00000101; // P303 = Output, P303= H, ATARI_P2_7

        attachInterrupt(digitalPinToInterrupt(3), put_atari_p2_normal, CHANGE);
        *(irq_ptr+PORT_IRQ1_VEC) = (uint32_t)put_atari_p2_normal;
        R_ICU->IRQCR[PORT_IRQ1_IRQ] = 0x02; // デジタルフィルタ無効, PCLKB, 両エッジ
      }
      break;
    case IF_MODE_JOYDRV:
      if(port == 0) {
        // UNO R4
        R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000101; // P014 = Output, P014= H, ATARI_P1_1
        R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000101; // P000 = Output, P000= H, ATARI_P1_2
        R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000101; // P001 = Output, P001= H, ATARI_P1_3
        R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000101; // P002 = Output, P002= H, ATARI_P1_4
        R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00010000; // P101 = Input, PULLUP = ON, ATARI_P1_6
        R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = 0b00010000; // P100 = Input, PULLUP = ON, ATARI_P1_7

        attachInterrupt(digitalPinToInterrupt(2), put_atari_p1_joydrv, CHANGE);
        *(irq_ptr+PORT_IRQ0_VEC) = (uint32_t)put_atari_p1_joydrv;
        R_ICU->IRQCR[PORT_IRQ0_IRQ] = 0x02; // デジタルフィルタ無効, PCLKB, 両エッジ
        joydrv_phase_cnt = 0;
        joydrv_connect_cnt = 0;
      }
      break;
    case IF_MODE_CYBER:
    case IF_MODE_XE1AP:
      if(port == 0) {
        // UNO R4
        R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000101; // P014 = Output, P014= H, ATARI_P1_1
        R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000101; // P000 = Output, P000= H, ATARI_P1_2
        R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000101; // P001 = Output, P001= H, ATARI_P1_3
        R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000101; // P002 = Output, P002= H, ATARI_P1_4
        R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P1_6
        R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = 0b00000101; // P100 = Output, P100= H, ATARI_P1_7

        attachInterrupt(digitalPinToInterrupt(2), put_atari_p1_cyber, FALLING);
        *(irq_ptr+PORT_IRQ0_VEC) = (uint32_t)put_atari_p1_cyber;
        R_ICU->IRQCR[PORT_IRQ0_IRQ] = 0x00; // デジタルフィルタ無効, PCLKB, ダウンエッジ
      }
      else {
        // UNO R4
        R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = 0b00000101; // P103 = Output, P103= H, ATARI_P2_1
        R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = 0b00000101; // P102 = Output, P102= H, ATARI_P2_2
        R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = 0b00000101; // P106 = Output, P106= H, ATARI_P2_3
        R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = 0b00000101; // P107 = Output, P107= H, ATARI_P2_4
        R_PFS->PORT[3].PIN[ 4].PmnPFS_BY = 0b00000101; // P304 = Output, P304= H, ATARI_P2_6
        R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = 0b00000101; // P303 = Output, P303= H, ATARI_P2_7

        attachInterrupt(digitalPinToInterrupt(3), put_atari_p2_cyber, FALLING);
        *(irq_ptr+PORT_IRQ1_VEC) = (uint32_t)put_atari_p2_cyber;
        R_ICU->IRQCR[PORT_IRQ1_IRQ] = 0x00; // デジタルフィルタ無効, PCLKB, ダウンエッジ
      }
      break;
    case IF_MODE_MD6B:
      if(port == 0) {
        // UNO R4
        R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000101; // P014 = Output, P014= H, ATARI_P1_1
        R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000101; // P000 = Output, P000= H, ATARI_P1_2
        R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000101; // P001 = Output, P001= H, ATARI_P1_3
        R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000101; // P002 = Output, P002= H, ATARI_P1_4
        R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P1_6
        R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = 0b00000101; // P100 = Output, P100= H, ATARI_P1_7

        attachInterrupt(digitalPinToInterrupt(2), put_atari_p1_md6b_s1, CHANGE);
        *(irq_ptr+PORT_IRQ0_VEC) = (uint32_t)put_atari_p1_md6b_s1;
        R_ICU->IRQCR[PORT_IRQ0_IRQ] = 0x02; // デジタルフィルタ無効, PCLKB, 両エッジ
      }
      else {
        // UNO R4
        R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = 0b00000101; // P103 = Output, P103= H, ATARI_P2_1
        R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = 0b00000101; // P102 = Output, P102= H, ATARI_P2_2
        R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = 0b00000101; // P106 = Output, P106= H, ATARI_P2_3
        R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = 0b00000101; // P107 = Output, P107= H, ATARI_P2_4
        R_PFS->PORT[3].PIN[ 4].PmnPFS_BY = 0b00000101; // P304 = Output, P304= H, ATARI_P2_6
        R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = 0b00000101; // P303 = Output, P303= H, ATARI_P2_7

        attachInterrupt(digitalPinToInterrupt(3), put_atari_p2_md6b_s1, CHANGE);
        *(irq_ptr+PORT_IRQ1_VEC) = (uint32_t)put_atari_p2_md6b_s1;
        R_ICU->IRQCR[PORT_IRQ1_IRQ] = 0x02; // デジタルフィルタ無効, PCLKB, 両エッジ
      }
      break;
    case IF_MODE_TWINPLUS2_AB:
    case IF_MODE_TWINPLUS2_BA:
    case IF_MODE_JAMMA_CPSF:
    case IF_MODE_JAMMA_MAGICAL:
    case IF_MODE_JAMMA_MD3B:
      // UNO R4
      R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = 0b00000101; // P103 = Output, P103= H, ATARI_P2_1
      R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = 0b00000101; // P102 = Output, P102= H, ATARI_P2_2
      R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = 0b00000101; // P106 = Output, P106= H, ATARI_P2_3
      R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = 0b00000101; // P107 = Output, P107= H, ATARI_P2_4
      R_PFS->PORT[3].PIN[ 4].PmnPFS_BY = 0b00000101; // P304 = Output, P304= H, ATARI_P2_6
      R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = 0b00000101; // P303 = Output, P303= H, ATARI_P2_7
      R_PFS->PORT[1].PIN[ 4].PmnPFS_HA = 0b0000000000010000; // P104 = Input, PULLUP = ON, INT = OFF, ATARI_P2_8

      if(port == 0) {
        // UNO R4
        R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000101; // P014 = Output, P014= H, ATARI_P1_1
        R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000101; // P000 = Output, P000= H, ATARI_P1_2
        R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000101; // P001 = Output, P001= H, ATARI_P1_3
        R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000101; // P002 = Output, P002= H, ATARI_P1_4
        R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P1_6
        R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = 0b00000101; // P100 = Output, P100= H, ATARI_P1_7
        R_PFS->PORT[1].PIN[ 5].PmnPFS_HA = 0b0000000000010000; // P105 = Input, PULLUP = ON, INT = OFF, ATARI_P1_8

      }
      break;
  }
  *(irq_pri+5) = 0x10;        // 外部ピン割り込みの優先順位変更(優先順位1)

  // UNO R4
  R_PFS->PORT[1].PIN[11].PmnPFS = 0x00000000; // P111 = Input
  return mode;
}

byte set_lplusad(byte port)
{
  byte work;

  work = (joydrv_snddata[port][3]|0b11110000);
  if(joydrv_snddata[port][5] < AD_CNV_DATA1) work &= 0b11111011; // 左
  else if(joydrv_snddata[port][5] > AD_CNV_DATA2) work &= 0b11110111; // 右
  if(joydrv_snddata[port][4] < AD_CNV_DATA1) work &= 0b11111110; // 上
  else if(joydrv_snddata[port][4] > AD_CNV_DATA2) work &= 0b11111101; // 下

  return work;
}

byte set_rplusad(byte port)
{
  byte work;

  work = ((joydrv_snddata[port][3]>>4)|0b11110000);
  if(joydrv_snddata[port][7] < AD_CNV_DATA1) work &= 0b11111011; // 左
  else if(joydrv_snddata[port][7] > AD_CNV_DATA2) work &= 0b11110111; // 右
  if(joydrv_snddata[port][6] < AD_CNV_DATA1) work &= 0b11111110; // 上
  else if(joydrv_snddata[port][6] > AD_CNV_DATA2) work &= 0b11111101; // 下

  return work;
}

void set_ataridata(byte port, byte mode)
{
  byte atari_work_SELL;
  byte atari_work_SELH;
  byte inf_btn_a,inf_btn_b,inf_btn_ar,inf_btn_br,inf_btn_r1,inf_btn_r2,inf_btn_l1,inf_btn_l2;

  if(mode == IF_MODE_ATARI_AB || mode == IF_MODE_TWINPLUS1_AB || mode == IF_MODE_TWINPLUS2_AB) {
    inf_btn_a  = 0x01;
    inf_btn_ar  = 0;
    inf_btn_b  = 0x02;
    inf_btn_br  = 1;
    inf_btn_r1 = 0x10;
    inf_btn_r2 = 0x20;
    inf_btn_l1 = 0x01;
    inf_btn_l2 = 0x02;
  }
  else {
    inf_btn_a  = 0x02;
    inf_btn_ar  = 1;
    inf_btn_b  = 0x01;
    inf_btn_br  = 0;
    inf_btn_r1 = 0x20;
    inf_btn_r2 = 0x10;
    inf_btn_l1 = 0x02;
    inf_btn_l2 = 0x01;
  }

  atari_work_SELL = set_lplusad(port);
  atari_work_SELH = set_rplusad(port);
  if(!(joydrv_snddata[port][0]&0x01)) // STARTボタン
    atari_work_SELL &= 0b11110011;
  if(!(joydrv_snddata[port][0]&0x02)) // SELECTボタン
    atari_work_SELL &= 0b11111100;

  if(!(joydrv_snddata[port][2]&inf_btn_a) || !(joydrv_snddata[port][1]&inf_btn_l1) || rensha_time[port][inf_btn_ar]<RENSHA_SPEED) // 1P Aボタン
    atari_work_SELL &= 0b11101111;
  if(!(joydrv_snddata[port][1]&inf_btn_r1)) // 2P Aボタン
    atari_work_SELH &= 0b11101111;
  if(!(joydrv_snddata[port][2]&inf_btn_b) || !(joydrv_snddata[port][1]&inf_btn_l2) || rensha_time[port][inf_btn_br]<RENSHA_SPEED) // 1P Bボタン
    atari_work_SELL &= 0b11011111;
  if(!(joydrv_snddata[port][1]&inf_btn_r2)) // 2P Bボタン
    atari_work_SELH &= 0b11011111;
  if(!(joydrv_snddata[port][2]&0x04) || rensha_time[port][2]<RENSHA_SPEED) // 1P Cボタン
    atari_work_SELL &= 0b11001111;

  if(mode == IF_MODE_ATARI_AB || mode == IF_MODE_ATARI_BA) {
    atari_work_SELL &= atari_work_SELH;
    atari_work_SELH = atari_work_SELL;
  }
  if(mode == IF_MODE_TWINPLUS1_AB || mode == IF_MODE_TWINPLUS1_BA) {
    if(!(atari_work_SELL&0b00010000)) atari_work_SELH &= 0b11101111;
    if(!(atari_work_SELL&0b00100000)) atari_work_SELH &= 0b11011111;
  }
  atari_data_SELL[port][0] = ((atari_work_SELL&0x01) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][1] = ((atari_work_SELL&0x02) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][2] = ((atari_work_SELL&0x04) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][3] = ((atari_work_SELL&0x08) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][4] = ((atari_work_SELL&0x10) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][5] = ((atari_work_SELL&0x20) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][0] = ((atari_work_SELH&0x01) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][1] = ((atari_work_SELH&0x02) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][2] = ((atari_work_SELH&0x04) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][3] = ((atari_work_SELH&0x08) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][4] = ((atari_work_SELH&0x10) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][5] = ((atari_work_SELH&0x20) ? 0b00000101 : 0b00000100);
}

void set_cpsfdata(byte port, byte mode)
{
  byte atari_work_SELL;
  byte atari_work_SELH;

  atari_work_SELL = set_lplusad(port);
  atari_work_SELL &= set_rplusad(port);
  atari_work_SELH = ((mode == IF_MODE_CPSF) ? 0b11111111 : 0b11111100);

  if(mode == IF_MODE_CPSF) {
    if(!(joydrv_snddata[port][0]&0x01)) // STARTボタン
      atari_work_SELH &= 0b11011111;
    if(!(joydrv_snddata[port][0]&0x02)) // SELECTボタン
      atari_work_SELH &= 0b11110111;
  }
  if(!(joydrv_snddata[port][2]&0x01)) // Aボタン
    atari_work_SELL &= ((mode == IF_MODE_CPSF) ? 0b11101111 : 0b11011111);
  if(!(joydrv_snddata[port][2]&0x02)) // Bボタン
    atari_work_SELL &= ((mode == IF_MODE_CPSF) ? 0b11011111 : 0b11101111);
  if(!(joydrv_snddata[port][1]&0x10) || !(joydrv_snddata[port][1]&0x02) || !(joydrv_snddata[port][2]&0x04)) // R1ボタンorL2ボタンorCボタン
    atari_work_SELH &= ((mode == IF_MODE_CPSF) ? 0b11111110 : 0b11110111);
  if(!(joydrv_snddata[port][2]&0x10)) // Xボタン
    atari_work_SELH &= ((mode == IF_MODE_CPSF) ? 0b11111011 : 0b11011111);
  if(!(joydrv_snddata[port][2]&0x20)) // Yボタン
    atari_work_SELH &= ((mode == IF_MODE_CPSF) ? 0b11111101 : 0b11101111);
  if(!(joydrv_snddata[port][1]&0x01) || !(joydrv_snddata[port][1]&0x20) || !(joydrv_snddata[port][2]&0x40)) // L1ボタンorR2ボタンorZボタン
    atari_work_SELH &= ((mode == IF_MODE_CPSF) ? 0b11101111 : 0b11111011);
  atari_data_SELL[port][0] = ((atari_work_SELL&0x01) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][1] = ((atari_work_SELL&0x02) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][2] = ((atari_work_SELL&0x04) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][3] = ((atari_work_SELL&0x08) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][4] = ((atari_work_SELL&0x10) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][5] = ((atari_work_SELL&0x20) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][0] = ((atari_work_SELH&0x01) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][1] = ((atari_work_SELH&0x02) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][2] = ((atari_work_SELH&0x04) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][3] = ((atari_work_SELH&0x08) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][4] = ((atari_work_SELH&0x10) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][5] = ((atari_work_SELH&0x20) ? 0b00000101 : 0b00000100);
}

void set_mddata(byte port,byte mode)
{
  byte atari_work_SELL;
  byte atari_work_SELH;
  byte md6b_work_SELL;
  byte md6b_work_SELH;

  atari_work_SELH = set_lplusad(port);
  atari_work_SELH &= set_rplusad(port);
  atari_work_SELL = (atari_work_SELH&0b11110011);
  md6b_work_SELH = 0b11111111;

  if(!(joydrv_snddata[port][0]&0x01)) // STARTボタン
    atari_work_SELL &= 0b11011111;
  if(!(joydrv_snddata[port][2]&0x01) || (rensha_time[port][0]<RENSHA_SPEED && (mode == IF_MODE_MD3B || mode == IF_MODE_JAMMA_MD3B))) // Aボタン
    atari_work_SELL &= 0b11101111;
  if(!(joydrv_snddata[port][2]&0x02) || (rensha_time[port][1]<RENSHA_SPEED && (mode == IF_MODE_MD3B || mode == IF_MODE_JAMMA_MD3B))) // Bボタン
    atari_work_SELH &= 0b11101111;
  if(!(joydrv_snddata[port][1]&0x10) || !(joydrv_snddata[port][1]&0x02) || !(joydrv_snddata[port][2]&0x04) || (rensha_time[port][2]<RENSHA_SPEED && (mode == IF_MODE_MD3B || mode == IF_MODE_JAMMA_MD3B))) // R1ボタンorL2ボタンorCボタン
    atari_work_SELH &= 0b11011111;

  if(!(joydrv_snddata[port][0]&0x02)) // SELECTボタン
    md6b_work_SELH &= 0b11110111;
  if(!(joydrv_snddata[port][2]&0x10)) // Xボタン
    md6b_work_SELH &= 0b11111011;
  if(!(joydrv_snddata[port][2]&0x20)) // Yボタン
    md6b_work_SELH &= 0b11111101;
  if(!(joydrv_snddata[port][1]&0x01) || !(joydrv_snddata[port][1]&0x20) || !(joydrv_snddata[port][2]&0x40)) // L1ボタンorR2ボタンorZボタン
    md6b_work_SELH &= 0b11111110;
  md6b_work_SELL = (atari_work_SELL&0b11110000);

  atari_data_SELL[port][0] = ((atari_work_SELL&0x01) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][1] = ((atari_work_SELL&0x02) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][2] = ((atari_work_SELL&0x04) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][3] = ((atari_work_SELL&0x08) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][4] = ((atari_work_SELL&0x10) ? 0b00000101 : 0b00000100);
  atari_data_SELL[port][5] = ((atari_work_SELL&0x20) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][0] = ((atari_work_SELH&0x01) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][1] = ((atari_work_SELH&0x02) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][2] = ((atari_work_SELH&0x04) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][3] = ((atari_work_SELH&0x08) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][4] = ((atari_work_SELH&0x10) ? 0b00000101 : 0b00000100);
  atari_data_SELH[port][5] = ((atari_work_SELH&0x20) ? 0b00000101 : 0b00000100);

  md6b_data_SELL[port][0] = ((md6b_work_SELL&0x01) ? 0b00000101 : 0b00000100);
  md6b_data_SELL[port][1] = ((md6b_work_SELL&0x02) ? 0b00000101 : 0b00000100);
  md6b_data_SELL[port][2] = ((md6b_work_SELL&0x04) ? 0b00000101 : 0b00000100);
  md6b_data_SELL[port][3] = ((md6b_work_SELL&0x08) ? 0b00000101 : 0b00000100);
  md6b_data_SELL[port][4] = ((md6b_work_SELL&0x10) ? 0b00000101 : 0b00000100);
  md6b_data_SELL[port][5] = ((md6b_work_SELL&0x20) ? 0b00000101 : 0b00000100);
  md6b_data_SELH[port][0] = ((md6b_work_SELH&0x01) ? 0b00000101 : 0b00000100);
  md6b_data_SELH[port][1] = ((md6b_work_SELH&0x02) ? 0b00000101 : 0b00000100);
  md6b_data_SELH[port][2] = ((md6b_work_SELH&0x04) ? 0b00000101 : 0b00000100);
  md6b_data_SELH[port][3] = ((md6b_work_SELH&0x08) ? 0b00000101 : 0b00000100);
  md6b_data_SELH[port][4] = ((md6b_work_SELH&0x10) ? 0b00000101 : 0b00000100);
  md6b_data_SELH[port][5] = ((md6b_work_SELH&0x20) ? 0b00000101 : 0b00000100);
}

void loop() {
  byte chg_mode;
  byte gpio_work1,gpio_work2;
  byte i;

  mtimes_now = millis();
  if(mtimes_now != mtimes_pre) {
    mode_time[0]++;
    mode_time[1]++;
    md6b_time[0]++;
    md6b_time[1]++;

    if(!(joydrv_snddata[0][2]&0x10)) { // Xボタン
      if(++rensha_time[0][0]>RENSHA_SPEED*2) rensha_time[0][0] = 0;
    }
    else rensha_time[0][0] = 254;

    if(!(joydrv_snddata[0][2]&0x20)) { // Yボタン
      if(++rensha_time[0][1]>RENSHA_SPEED*2) rensha_time[0][1] = 0;
    }
    else rensha_time[0][1] = 254;

    if(((cnv_mode[0] == IF_MODE_MD3B || cnv_mode[0] == IF_MODE_JAMMA_MD3B) && (!(joydrv_snddata[0][1]&0x01) || !(joydrv_snddata[0][1]&0x20))) || !(joydrv_snddata[0][2]&0x40)) { // L1ボタン or R2ボタン or Zボタン
      if(++rensha_time[0][2]>RENSHA_SPEED*2) rensha_time[0][2] = 0;
    }
    else rensha_time[0][2] = 254;

    if(!(joydrv_snddata[1][2]&0x10)) { // Xボタン
      if(++rensha_time[1][0]>RENSHA_SPEED*2) rensha_time[1][0] = 0;
    }
    else rensha_time[1][0] = 254;

    if(!(joydrv_snddata[1][2]&0x20)) { // Yボタン
      if(++rensha_time[1][1]>RENSHA_SPEED*2) rensha_time[1][1] = 0;
    }
    else rensha_time[1][1] = 254;

    if(((cnv_mode[1] == IF_MODE_MD3B || cnv_mode[1] == IF_MODE_JAMMA_MD3B) && (!(joydrv_snddata[1][1]&0x01) || !(joydrv_snddata[1][1]&0x20))) || !(joydrv_snddata[1][2]&0x40)) { // L1ボタン or R2ボタン or Zボタン
      if(++rensha_time[1][2]>RENSHA_SPEED*2) rensha_time[1][2] = 0;
    }
    else rensha_time[1][2] = 254;

    output_time[0]++;
    output_time[1]++;
    if(gpio_cnt >= 0) gpio_cnt++;
    usbtask_time++;
    if(joydrv_phase_cnt>0) {
      if(++joydrv_connect_cnt>JOYDRV_WAIT) {
        joydrv_phase_cnt=0;
        joydrv_connect_cnt=0;
        // UNO R4
        R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000101;
        R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000101;
        R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000101;
        R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000101;
        R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00010000; // P101 = Input, PULLUP = ON, ATARI_P0_6
      }
    }
    mtimes_pre = millis();
  }

  if(usbtask_time) {
    Usb.Task();
    usbtask_time = 0;
  }

  if(md6b_time[0] >= MD6B_RESETTIME) {
    md6b_cnt[0] = 0;
    md6b_time[0] = 0;
  }   
  if(md6b_time[1] >= MD6B_RESETTIME) {
    md6b_cnt[1] = 0;
    md6b_time[1] = 0;
  }

  for(i=0;i<2;i++) {
    if(!(joydrv_snddata[i][0]&0x01)) { //STARTボタン
      chg_mode = cnv_mode[i];
      if(!(joydrv_snddata[i][2]&0x01)) { // Aボタン
        if(modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_ATARI_AB : IF_MODE_ATARI_BA);
      }
      else if(!(joydrv_snddata[i][2]&0x02)) { // Bボタン
        if(modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_TWINPLUS1_AB : IF_MODE_TWINPLUS1_BA);
      }
      else if(!(joydrv_snddata[i][2]&0x10)) { // Xボタン
        if(modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_CYBER : IF_MODE_XE1AP);
      }
      else if(!(joydrv_snddata[i][2]&0x20)) { // Yボタン
        if(i==1 || modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_JOYDRV : IF_MODE_JAMMA_MD3B);
      }
      else if(!(joydrv_snddata[i][3]&0x01) || joydrv_snddata[i][4] < AD_CNV_DATA1) { // 十字キー上 or 左レバー上
        if(i==1 || modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_TWINPLUS2_AB : IF_MODE_TWINPLUS2_BA);
      }
      else if(!(joydrv_snddata[i][3]&0x02) || joydrv_snddata[i][4] > AD_CNV_DATA2) { // 十字キー下 or 左レバー下
        if(modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_CPSF : IF_MODE_MAGICAL);
      }
      else if(!(joydrv_snddata[i][3]&0x04) || joydrv_snddata[i][5] < AD_CNV_DATA1) { // 十字キー左 or 左レバー左
        if(modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_MD3B : IF_MODE_MD6B);
      }
      else if(!(joydrv_snddata[i][3]&0x08) || joydrv_snddata[i][5] > AD_CNV_DATA2) { // 十字キー右 or 左レバー右
        if(i==1 || modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_JAMMA_CPSF : IF_MODE_JAMMA_MAGICAL);
      }
      else if(!(joydrv_snddata[i][0]&0x02)) // SELECTボタン
        chg_mode = 255;
      else mode_time[i] = 0;
      if(mode_time[i] >= MODE_CHG_TIME) {
        flg_chg = 1;
        if(chg_mode == 255) {
          modechg_sts[i] ^= 0b00000001;
          EEPROM.write(i+2, modechg_sts[i]);
        }        
        else {
          cnv_mode[i] = chg_mode;
          EEPROM.write(i, chg_mode);
        }
      }
    }
    else mode_time[i] = 0;
  }

  gpio_work1 = gpio_work2 = Usb.gpioRd();
  gpio_work1 &= 0x0f;
  gpio_work2 = (gpio_work2>>4)&0x0f;

  if(gpio_sw1 != gpio_work1) {
    gpio_sw1 = gpio_work1;
    if(modechg_sts[0] == 1) gpio_cnt = 0;
  }
  if(gpio_sw2 != gpio_work2) {
    gpio_sw2 = gpio_work2;
    if(modechg_sts[1] == 1) gpio_cnt = 0;
  }
  if(gpio_cnt >= SW_CHANGE_MARGIN) {
    gpio_cnt = -1;
    flg_chg = 1;
  }

  if(flg_chg == 1) {
    if(modechg_sts[0] == 1) {
      cnv_mode[0] = (gpio_sw1^0b00001111);
      EEPROM.write(0, cnv_mode[0]);
    }
    if(modechg_sts[1] == 1) {
      cnv_mode[1] = (gpio_sw2^0b00001111);
      if(cnv_mode[1] == IF_MODE_JOYDRV || cnv_mode[1] == IF_MODE_TWINPLUS2_AB || cnv_mode[1] == IF_MODE_JAMMA_CPSF || cnv_mode[1] == IF_MODE_JAMMA_MD3B || cnv_mode[1] == IF_MODE_TWINPLUS2_BA || cnv_mode[1] == IF_MODE_JAMMA_MAGICAL)
        cnv_mode[1] = IF_MODE_ATARI_AB;
      EEPROM.write(1, cnv_mode[1]);
    }

    detachInterrupt(digitalPinToInterrupt(2));
    detachInterrupt(digitalPinToInterrupt(3));

    SPI.end();
    gpio_work1 = set_cnv_mode(0);
    gpio_work2 = set_cnv_mode(1);
    SPI.begin();

    lcd.setCursor(0, 0);
    lcd_mode_data[gpio_work1][1]= '1';
    lcd_mode_data[gpio_work1][3]= (modechg_sts[0] == 0 ? 'C' : 'S');
    lcd.print(lcd_mode_data[gpio_work1]);
    lcd.setCursor(0, 1);
    lcd_mode_data[gpio_work2][1]= '2';
    lcd_mode_data[gpio_work2][3]= (modechg_sts[1] == 0 ? 'C' : 'S');
    lcd.print(lcd_mode_data[gpio_work2]);

    flg_chg = 0;
    mode_time[0] = mode_time[1] = 0; /* モード変更時間の初期化 */ 
    mtimes_pre = millis();
  }

  // ATARI_P1側処理
  switch(cnv_mode[0]) {
    case IF_MODE_ATARI_AB:
    case IF_MODE_ATARI_BA:
      set_ataridata(0, cnv_mode[0]);
      put_atari_p1_normal();
      break;
    case IF_MODE_TWINPLUS1_AB:
    case IF_MODE_TWINPLUS1_BA:
      set_ataridata(0, cnv_mode[0]);
      goto IF_MODE_P1;
    case IF_MODE_CPSF:
    case IF_MODE_MAGICAL:
      set_cpsfdata(0, cnv_mode[0]);
      goto IF_MODE_P1;
    case IF_MODE_MD3B:
      set_mddata(0, IF_MODE_MD3B);
      IF_MODE_P1:
      if(output_time[0] >= OUTLIMIT_TIME) put_atari_p1_normal();
      break;
    case IF_MODE_MD6B:
      set_mddata(0, IF_MODE_MD6B);
      if(output_time[0] >= OUTLIMIT_TIME) put_atari_p1_md6b_s2();
      break;
    case IF_MODE_JAMMA_MD3B:
      set_mddata(0, IF_MODE_JAMMA_MD3B);
      goto IF_MODE_JAMMA_COM;
    case IF_MODE_JAMMA_MAGICAL:
      set_cpsfdata(0, IF_MODE_MAGICAL);
      goto IF_MODE_JAMMA_COM;
    case IF_MODE_JAMMA_CPSF:
      set_cpsfdata(0, IF_MODE_CPSF);
      goto IF_MODE_JAMMA_COM;
    case IF_MODE_TWINPLUS2_AB:
    case IF_MODE_TWINPLUS2_BA:
      set_ataridata(0, cnv_mode[0]);
      IF_MODE_JAMMA_COM:
      for(i=0;i<6;i++) {
        atari_data_SELH[1][i]=atari_data_SELL[1][i]=atari_data_SELH[0][i];
        atari_data_SELH[0][i]=atari_data_SELL[0][i];
      }
      put_atari_p1_normal();
      put_atari_p2_normal();
      break;
  } // CYBER,JOYDRVIFは処理なし

  // ATARI_P2側処理
  if(!(cnv_mode[0] == IF_MODE_JOYDRV || cnv_mode[0] == IF_MODE_TWINPLUS2_AB || cnv_mode[0] == IF_MODE_JAMMA_CPSF || cnv_mode[0] == IF_MODE_JAMMA_MD3B || cnv_mode[0] == IF_MODE_TWINPLUS2_BA || cnv_mode[0] == IF_MODE_JAMMA_MAGICAL)) {
    switch(cnv_mode[1]) {
      case IF_MODE_ATARI_AB:
      case IF_MODE_ATARI_BA:
        set_ataridata(1, cnv_mode[1]);
        put_atari_p2_normal();
        break;
      case IF_MODE_TWINPLUS1_AB:
      case IF_MODE_TWINPLUS1_BA:
        set_ataridata(1, cnv_mode[1]);
        goto IF_MODE_P2;
      case IF_MODE_CPSF:
      case IF_MODE_MAGICAL:
        set_cpsfdata(1, cnv_mode[1]);
        goto IF_MODE_P2;
      case IF_MODE_MD3B:
        set_mddata(1, IF_MODE_MD3B);
        IF_MODE_P2:
        if(output_time[1] >= OUTLIMIT_TIME) put_atari_p2_normal();
        break;
      case IF_MODE_MD6B:
        set_mddata(1, IF_MODE_MD6B);
        if(output_time[1] >= OUTLIMIT_TIME) put_atari_p2_md6b_s2();
        break;
    }
  }
}

void put_atari_p1_normal()
{
  R_ICU->IELSR[PORT_IRQ0_ISEL] = 1; // PORT_IRQ0, 割り込み要求クリア
  if(R_PFS->PORT[1].PIN[ 5].PmnPFS_BY&0x02) {
    R_PFS->PORT[0].PIN[14].PmnPFS_BY = atari_data_SELH[0][0];
    R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = atari_data_SELH[0][1];
    R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = atari_data_SELH[0][2];
    R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = atari_data_SELH[0][3];
    R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = atari_data_SELH[0][4];
    R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = atari_data_SELH[0][5];
  }
  else {
    R_PFS->PORT[0].PIN[14].PmnPFS_BY = atari_data_SELL[0][0];
    R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = atari_data_SELL[0][1];
    R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = atari_data_SELL[0][2];
    R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = atari_data_SELL[0][3];
    R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = atari_data_SELL[0][4];
    R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = atari_data_SELL[0][5];
  }
  output_time[0] = 0;
//  R_ICU->IELSR[PORT_IRQ0_ISEL] = 1; // PORT_IRQ0, 割り込み要求クリア
}

void put_atari_p2_normal()
{
  R_ICU->IELSR[PORT_IRQ1_ISEL] = 2; // PORT_IRQ1, 割り込み要求クリア
  if(R_PFS->PORT[1].PIN[ 4].PmnPFS_BY&0x02) {
    R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = atari_data_SELH[1][0];
    R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = atari_data_SELH[1][1];
    R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = atari_data_SELH[1][2];
    R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = atari_data_SELH[1][3];
    R_PFS->PORT[3].PIN[ 4].PmnPFS_BY = atari_data_SELH[1][4];
    R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = atari_data_SELH[1][5];
  }
  else {
    R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = atari_data_SELL[1][0];
    R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = atari_data_SELL[1][1];
    R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = atari_data_SELL[1][2];
    R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = atari_data_SELL[1][3];
    R_PFS->PORT[3].PIN[ 4].PmnPFS_BY = atari_data_SELL[1][4];
    R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = atari_data_SELL[1][5];
  }
  output_time[1] = 0;
//  R_ICU->IELSR[PORT_IRQ1_ISEL] = 2; // PORT_IRQ1, 割り込み要求クリア
}

void put_atari_p1_md6b_s1()
{
  R_ICU->IELSR[PORT_IRQ0_ISEL] = 1; // PORT_IRQ0, 割り込み要求クリア
  if(!(R_PFS->PORT[1].PIN[ 5].PmnPFS_BY&0x02)) md6b_cnt[0]++;
  put_atari_p1_md6b_s2();
  md6b_time[0] = 0;
//  R_ICU->IELSR[PORT_IRQ0_ISEL] = 1; // PORT_IRQ0, 割り込み要求クリア
}

void put_atari_p1_md6b_s2()
{
  if(R_PFS->PORT[1].PIN[ 5].PmnPFS_BY&0x02) {
    if(md6b_cnt[0] == 3) {
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = md6b_data_SELH[0][0];
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = md6b_data_SELH[0][1];
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = md6b_data_SELH[0][2];
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = md6b_data_SELH[0][3];
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = md6b_data_SELH[0][4];
      R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = md6b_data_SELH[0][5];
    }
    else {
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = atari_data_SELH[0][0];
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = atari_data_SELH[0][1];
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = atari_data_SELH[0][2];
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = atari_data_SELH[0][3];
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = atari_data_SELH[0][4];
      R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = atari_data_SELH[0][5];
    }
  }
  else {
    if(md6b_cnt[0] == 3) {
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = md6b_data_SELL[0][0];
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = md6b_data_SELL[0][1];
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = md6b_data_SELL[0][2];
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = md6b_data_SELL[0][3];
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = md6b_data_SELL[0][4];
      R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = md6b_data_SELL[0][5];
    }
    else {
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = atari_data_SELL[0][0];
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = atari_data_SELL[0][1];
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = atari_data_SELL[0][2];
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = atari_data_SELL[0][3];
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = atari_data_SELL[0][4];
      R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = atari_data_SELL[0][5];
    }
  }
  output_time[0] = 0;
}

void put_atari_p2_md6b_s1()
{
  R_ICU->IELSR[PORT_IRQ1_ISEL] = 2; // PORT_IRQ1, 割り込み要求クリア
  if(!(R_PFS->PORT[1].PIN[ 4].PmnPFS_BY&0x02)) md6b_cnt[1]++;
  put_atari_p2_md6b_s2();
  md6b_time[1] = 0;
//  R_ICU->IELSR[PORT_IRQ1_ISEL] = 2; // PORT_IRQ1, 割り込み要求クリア
}

void put_atari_p2_md6b_s2()
{
  if(R_PFS->PORT[1].PIN[ 4].PmnPFS_BY&0x02) {
    if(md6b_cnt[1] == 3) {
      R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = md6b_data_SELH[1][0];
      R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = md6b_data_SELH[1][1];
      R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = md6b_data_SELH[1][2];
      R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = md6b_data_SELH[1][3];
      R_PFS->PORT[3].PIN[ 4].PmnPFS_BY = md6b_data_SELH[1][4];
      R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = md6b_data_SELH[1][5];
    }
    else {
      R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = atari_data_SELH[1][0];
      R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = atari_data_SELH[1][1];
      R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = atari_data_SELH[1][2];
      R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = atari_data_SELH[1][3];
      R_PFS->PORT[3].PIN[ 4].PmnPFS_BY = atari_data_SELH[1][4];
      R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = atari_data_SELH[1][5];
    }
  }
  else {
    if(md6b_cnt[1] == 3) {
      R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = md6b_data_SELL[1][0];
      R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = md6b_data_SELL[1][1];
      R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = md6b_data_SELL[1][2];
      R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = md6b_data_SELL[1][3];
      R_PFS->PORT[3].PIN[ 4].PmnPFS_BY = md6b_data_SELL[1][4];
      R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = md6b_data_SELL[1][5];
    }
    else {
      R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = atari_data_SELL[1][0];
      R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = atari_data_SELL[1][1];
      R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = atari_data_SELL[1][2];
      R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = atari_data_SELL[1][3];
      R_PFS->PORT[3].PIN[ 4].PmnPFS_BY = atari_data_SELL[1][4];
      R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = atari_data_SELL[1][5];
    }
  }
  output_time[1] = 0;
}

void put_atari_p1_cyber()
{
  byte i,j;
  byte cyber_data[12];

  // サイバースティックデータ送信
  // CYBER STICK
  //  0 : A+A' B+B' C D
  //  1 : E1 E2 F(START) G(SELECT)
  //  2 : 左上下 上位4bit
  //  3 : 左左右 上位4bit
  //  4 : 右上下 上位4bit
  //  5 : 右左右 上位4bit
  //  6 : 左上下 下位4bit
  //  7 : 左左右 下位4bit
  //  8 : 右上下 下位4bit
  //  9 : 右左右 下位4bit
  // 10 : A B A' B'
  // 11 : 1 1 1 1

  // XE-1AP
  //  0 : E1 E2 F(START) G(SELECT)
  //  1 : A+A' B+B' C D
  //  2 : 左左右 上位4bit
  //  3 : 左上下 上位4bit
  //  4 : 右左右 上位4bit
  //  5 : 右上下 上位4bit
  //  6 : 左左右 下位4bit
  //  7 : 左上下 下位4bit
  //  8 : 右左右 下位4bit
  //  9 : 右上下 下位4bit
  // 10 : E1 E2 E1 E2
  // 11 : 1 1 1 1
  if(cnv_mode[0]==IF_MODE_CYBER) {
    // 右左右アナログ
    cyber_data[3] = (joydrv_snddata[0][7]>>4);
    cyber_data[7] = joydrv_snddata[0][7];
    // 右上下アナログ
    cyber_data[2] = (joydrv_snddata[0][6]>>4);
    cyber_data[6] = joydrv_snddata[0][6];
    // 左左右アナログ
    cyber_data[5] = (joydrv_snddata[0][5]>>4);
    cyber_data[9] = joydrv_snddata[0][5];
    // 左上下アナログ
    cyber_data[4] = (joydrv_snddata[0][4]>>4);
    cyber_data[8] = joydrv_snddata[0][4];
    // Bボタン(A')
    if(joydrv_snddata[0][2] & 0x02) {
      cyber_data[ 0] |= 0x08;
      cyber_data[10] |= 0x02;
    }
    else {
      cyber_data[ 0] &= 0xf7;
      cyber_data[10] &= 0xfd;
    }
    // Aボタン(B')
    if(joydrv_snddata[0][2] & 0x01) {
      cyber_data[ 0] |= 0x04;
      cyber_data[10] |= 0x01;
    }
    else {
      cyber_data[ 0] &= 0xfb;
      cyber_data[10] &= 0xfe;
    }
    // R1ボタン(A)
    if(joydrv_snddata[0][1] & 0x10) {
      cyber_data[ 0] |= 0x08;
      cyber_data[10] |= 0x08;
    }
    else {
      cyber_data[ 0] &= 0xf7;
      cyber_data[10] &= 0xf7;
    }
    // R2ボタン(B)
    if(joydrv_snddata[0][1] & 0x20) {
      cyber_data[ 0] |= 0x04;
      cyber_data[10] |= 0x04;
    }
    else {
      cyber_data[ 0] &= 0xfb;
      cyber_data[10] &= 0xfb;
    }
    // Xボタン(E1)
    if(joydrv_snddata[0][2] & 0x10) cyber_data[ 1] |= 0x08;
    else cyber_data[ 1] &= 0xf7;
    // Yボタン(E2)
    if(joydrv_snddata[0][2] & 0x20) cyber_data[ 1] |= 0x04;
    else cyber_data[ 1] &= 0xfb;
    // L1ボタン(C)
    if(joydrv_snddata[0][1] & 0x01) cyber_data[ 0] |= 0x02;
    else cyber_data[ 0] &= 0xfd;
    // L2ボタン(D)
    if(joydrv_snddata[0][1] & 0x02) cyber_data[ 0] |= 0x01;
    else cyber_data[ 0] &= 0xfe;
    // STARTボタン
    if(joydrv_snddata[0][0] & 0x01) cyber_data[ 1] |= 0x02;
    else cyber_data[ 1] &= 0xfd;
    // SELECTボタン
    if(joydrv_snddata[0][0] & 0x02) cyber_data[ 1] |= 0x01;
    else cyber_data[ 1] &= 0xfe;
  }
  else {
    // 右左右アナログ
    cyber_data[2] = (joydrv_snddata[0][7]>>4);
    cyber_data[6] = joydrv_snddata[0][7];
    // 右上下アナログ
    cyber_data[3] = (joydrv_snddata[0][6]>>4);
    cyber_data[7] = joydrv_snddata[0][6];
    // 左左右アナログ
    cyber_data[4] = (joydrv_snddata[0][5]>>4);
    cyber_data[8] = joydrv_snddata[0][5];
    // 左上下アナログ
    cyber_data[5] = (joydrv_snddata[0][4]>>4);
    cyber_data[9] = joydrv_snddata[0][4];
    // Bボタン(A')
    if(joydrv_snddata[0][2] & 0x02) {
      cyber_data[ 1] |= 0x08;
      cyber_data[10] |= 0x02;
    }
    else {
      cyber_data[ 1] &= 0xf7;
      cyber_data[10] &= 0xfd;
    }
    // Aボタン(B')
    if(joydrv_snddata[0][2] & 0x01) {
      cyber_data[ 1] |= 0x04;
      cyber_data[10] |= 0x01;
    }
    else {
      cyber_data[ 1] &= 0xfb;
      cyber_data[10] &= 0xfe;
    }
    // R1ボタン(A)
    if(joydrv_snddata[0][1] & 0x10) {
      cyber_data[ 1] |= 0x08;
      cyber_data[10] |= 0x08;
    }
    else {
      cyber_data[ 1] &= 0xf7;
      cyber_data[10] &= 0xf7;
    }
    // R2ボタン(B)
    if(joydrv_snddata[0][1] & 0x20) {
      cyber_data[ 1] |= 0x04;
      cyber_data[10] |= 0x04;
    }
    else {
      cyber_data[ 1] &= 0xfb;
      cyber_data[10] &= 0xfb;
    }
    // Xボタン(E1)
    if(joydrv_snddata[0][2] & 0x10) cyber_data[ 0] |= 0x08;
    else cyber_data[ 0] &= 0xf7;
    // Yボタン(E2)
    if(joydrv_snddata[0][2] & 0x20) cyber_data[ 0] |= 0x04;
    else cyber_data[ 0] &= 0xfb;
    // L1ボタン(C)
    if(joydrv_snddata[0][1] & 0x01) cyber_data[ 1] |= 0x02;
    else cyber_data[ 1] &= 0xfd;
    // L2ボタン(D)
    if(joydrv_snddata[0][1] & 0x02) cyber_data[ 1] |= 0x01;
    else cyber_data[ 1] &= 0xfe;
    // STARTボタン
    if(joydrv_snddata[0][0] & 0x01) cyber_data[ 0] |= 0x02;
    else cyber_data[ 0] &= 0xfd;
    // SELECTボタン
    if(joydrv_snddata[0][0] & 0x02) cyber_data[ 0] |= 0x01;
    else cyber_data[ 0] &= 0xfe;
  }
  cyber_data[11] = 0x0f;

  delayMicroseconds(15);   // 7+15us

  j = 12;
  i = 0;
  while(i<12) {
    R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((cyber_data[i]&0x01) ? 0b00000101 : 0b00000100);
    R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((cyber_data[i]&0x02) ? 0b00000101 : 0b00000100);
    R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((cyber_data[i]&0x04) ? 0b00000101 : 0b00000100);
    R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((cyber_data[i++]&0x08) ? 0b00000101 : 0b00000100);
    // LH=L,ACK=H
    R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100;
    R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = 0b00000101;
    delayMicroseconds(22);   // 22us

    // ACK=L
    R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = 0b00000100;
    if(!(R_PFS->PORT[1].PIN[ 5].PmnPFS_BY&0x02)) {
      switch(j) {
        case 12:
          j=24;
          break;
        case 24:
          j=48;
          break;
        case 48:
          j=72;
          break;
      }
    }
    delayMicroseconds((unsigned int)cyber_speed1);  // 12us,24us,48us,72us

    R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((cyber_data[i]&0x01) ? 0b00000101 : 0b00000100);
    R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((cyber_data[i]&0x02) ? 0b00000101 : 0b00000100);
    R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((cyber_data[i]&0x04) ? 0b00000101 : 0b00000100);
    R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((cyber_data[i++]&0x08) ? 0b00000101 : 0b00000100);
    // LH=H,ACK=H
    R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101;
    R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = 0b00000101;
    delayMicroseconds((unsigned int)(cyber_speed1 == 12 ? 4 : 8));   // 4 or 8us

    // ACK=L
    R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = 0b00000100;
    delayMicroseconds((unsigned int)cyber_speed1);  // 12us,24us,48us,72us
  }
  R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000101;
  R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000101;
  R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000101;
  R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000101;
  R_PFS->PORT[1].PIN[ 0].PmnPFS_BY = 0b00000101;
  cyber_speed1 = j;
  R_ICU->IELSR[PORT_IRQ0_ISEL] = 1; // PORT_IRQ0, 割り込み要求クリア
}

void put_atari_p2_cyber()
{
  byte i,j;
  byte cyber_data[12];

  // サイバースティックデータ送信
  // CYBER STICK
  //  0 : A+A' B+B' C D
  //  1 : E1 E2 F(START) G(SELECT)
  //  2 : 左上下 上位4bit
  //  3 : 左左右 上位4bit
  //  4 : 右上下 上位4bit
  //  5 : 右左右 上位4bit
  //  6 : 左上下 下位4bit
  //  7 : 左左右 下位4bit
  //  8 : 右上下 下位4bit
  //  9 : 右左右 下位4bit
  // 10 : A B A' B'
  // 11 : 1 1 1 1

  // XE-1AP
  //  0 : E1 E2 F(START) G(SELECT)
  //  1 : A+A' B+B' C D
  //  2 : 左左右 上位4bit
  //  3 : 左上下 上位4bit
  //  4 : 右左右 上位4bit
  //  5 : 右上下 上位4bit
  //  6 : 左左右 下位4bit
  //  7 : 左上下 下位4bit
  //  8 : 右左右 下位4bit
  //  9 : 右上下 下位4bit
  // 10 : E1 E2 E1 E2
  // 11 : 1 1 1 1
  if(cnv_mode[1]==IF_MODE_CYBER) {
    // 右左右アナログ
    cyber_data[3] = (joydrv_snddata[1][7]>>4);
    cyber_data[7] = joydrv_snddata[1][7];
    // 右上下アナログ
    cyber_data[2] = (joydrv_snddata[1][6]>>4);
    cyber_data[6] = joydrv_snddata[1][6];
    // 左左右アナログ
    cyber_data[5] = (joydrv_snddata[1][5]>>4);
    cyber_data[9] = joydrv_snddata[1][5];
    // 左上下アナログ
    cyber_data[4] = (joydrv_snddata[1][4]>>4);
    cyber_data[8] = joydrv_snddata[1][4];
    // Bボタン(A')
    if(joydrv_snddata[1][2] & 0x02) {
      cyber_data[ 0] |= 0x08;
      cyber_data[10] |= 0x02;
    }
    else {
      cyber_data[ 0] &= 0xf7;
      cyber_data[10] &= 0xfd;
    }
    // Aボタン(B')
    if(joydrv_snddata[1][2] & 0x01) {
      cyber_data[ 0] |= 0x04;
      cyber_data[10] |= 0x01;
    }
    else {
      cyber_data[ 0] &= 0xfb;
      cyber_data[10] &= 0xfe;
    }
    // R1ボタン(A)
    if(joydrv_snddata[1][1] & 0x10) {
      cyber_data[ 0] |= 0x08;
      cyber_data[10] |= 0x08;
    }
    else {
      cyber_data[ 0] &= 0xf7;
      cyber_data[10] &= 0xf7;
    }
    // R2ボタン(B)
    if(joydrv_snddata[1][1] & 0x20) {
      cyber_data[ 0] |= 0x04;
      cyber_data[10] |= 0x04;
    }
    else {
      cyber_data[ 0] &= 0xfb;
      cyber_data[10] &= 0xfb;
    }
    // Xボタン(E1)
    if(joydrv_snddata[1][2] & 0x10) cyber_data[ 1] |= 0x08;
    else cyber_data[ 1] &= 0xf7;
    // Yボタン(E2)
    if(joydrv_snddata[1][2] & 0x20) cyber_data[ 1] |= 0x04;
    else cyber_data[ 1] &= 0xfb;
    // L1ボタン(C)
    if(joydrv_snddata[1][1] & 0x01) cyber_data[ 0] |= 0x02;
    else cyber_data[ 0] &= 0xfd;
    // L2ボタン(D)
    if(joydrv_snddata[1][1] & 0x02) cyber_data[ 0] |= 0x01;
    else cyber_data[ 0] &= 0xfe;
    // STARTボタン
    if(joydrv_snddata[1][0] & 0x01) cyber_data[ 1] |= 0x02;
    else cyber_data[ 1] &= 0xfd;
    // SELECTボタン
    if(joydrv_snddata[1][0] & 0x02) cyber_data[ 1] |= 0x01;
    else cyber_data[ 1] &= 0xfe;
  }
  else {
    // 右左右アナログ
    cyber_data[2] = (joydrv_snddata[1][7]>>4);
    cyber_data[6] = joydrv_snddata[1][7];
    // 右上下アナログ
    cyber_data[3] = (joydrv_snddata[1][6]>>4);
    cyber_data[7] = joydrv_snddata[1][6];
    // 左左右アナログ
    cyber_data[4] = (joydrv_snddata[1][5]>>4);
    cyber_data[8] = joydrv_snddata[1][5];
    // 左上下アナログ
    cyber_data[5] = (joydrv_snddata[1][4]>>4);
    cyber_data[9] = joydrv_snddata[1][4];
    // Bボタン(A')
    if(joydrv_snddata[1][2] & 0x02) {
      cyber_data[ 1] |= 0x08;
      cyber_data[10] |= 0x02;
    }
    else {
      cyber_data[ 1] &= 0xf7;
      cyber_data[10] &= 0xfd;
    }
    // Aボタン(B')
    if(joydrv_snddata[1][2] & 0x01) {
      cyber_data[ 1] |= 0x04;
      cyber_data[10] |= 0x01;
    }
    else {
      cyber_data[ 1] &= 0xfb;
      cyber_data[10] &= 0xfe;
    }
    // R1ボタン(A)
    if(joydrv_snddata[1][1] & 0x10) {
      cyber_data[ 1] |= 0x08;
      cyber_data[10] |= 0x08;
    }
    else {
      cyber_data[ 1] &= 0xf7;
      cyber_data[10] &= 0xf7;
    }
    // R2ボタン(B)
    if(joydrv_snddata[1][1] & 0x20) {
      cyber_data[ 1] |= 0x04;
      cyber_data[10] |= 0x04;
    }
    else {
      cyber_data[ 1] &= 0xfb;
      cyber_data[10] &= 0xfb;
    }
    // Xボタン(E1)
    if(joydrv_snddata[1][2] & 0x10) cyber_data[ 0] |= 0x08;
    else cyber_data[ 0] &= 0xf7;
    // Yボタン(E2)
    if(joydrv_snddata[1][2] & 0x20) cyber_data[ 0] |= 0x04;
    else cyber_data[ 0] &= 0xfb;
    // L1ボタン(C)
    if(joydrv_snddata[1][1] & 0x01) cyber_data[ 1] |= 0x02;
    else cyber_data[ 1] &= 0xfd;
    // L2ボタン(D)
    if(joydrv_snddata[1][1] & 0x02) cyber_data[ 1] |= 0x01;
    else cyber_data[ 1] &= 0xfe;
    // STARTボタン
    if(joydrv_snddata[1][0] & 0x01) cyber_data[ 0] |= 0x02;
    else cyber_data[ 0] &= 0xfd;
    // SELECTボタン
    if(joydrv_snddata[1][0] & 0x02) cyber_data[ 0] |= 0x01;
    else cyber_data[ 0] &= 0xfe;
  }
  cyber_data[11] = 0x0f;

  delayMicroseconds(15);   // 7+15us

  j = 12;
  i = 0;
  while(i<12) {
    R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = ((cyber_data[i]&0x01) ? 0b00000101 : 0b00000100);
    R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = ((cyber_data[i]&0x02) ? 0b00000101 : 0b00000100);
    R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = ((cyber_data[i]&0x04) ? 0b00000101 : 0b00000100);
    R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = ((cyber_data[i++]&0x08) ? 0b00000101 : 0b00000100);
    // LH=L,ACK=H
    R_PFS->PORT[3].PIN[ 4].PmnPFS_BY = 0b00000100;
    R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = 0b00000101;
    delayMicroseconds(22);   // 22us

    // ACK=L
    R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = 0b00000100;
    if(!(R_PFS->PORT[1].PIN[ 4].PmnPFS_BY&0x02)) {
      switch(j) {
        case 12:
          j=24;
          break;
        case 24:
          j=48;
          break;
        case 48:
          j=72;
          break;
      }
    }
    delayMicroseconds((unsigned int)cyber_speed2);  // 12us,24us,48us,72us

    R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = ((cyber_data[i]&0x01) ? 0b00000101 : 0b00000100);
    R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = ((cyber_data[i]&0x02) ? 0b00000101 : 0b00000100);
    R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = ((cyber_data[i]&0x04) ? 0b00000101 : 0b00000100);
    R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = ((cyber_data[i++]&0x08) ? 0b00000101 : 0b00000100);
    // LH=H,ACK=H
    R_PFS->PORT[3].PIN[ 4].PmnPFS_BY = 0b00000101;
    R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = 0b00000101;
    delayMicroseconds((unsigned int)(cyber_speed2 == 12 ? 4 : 8));   // 4 or 8us

    // ACK=L
    R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = 0b00000100;
    delayMicroseconds((unsigned int)cyber_speed2);  // 12us,24us,48us,72us
  }
  R_PFS->PORT[1].PIN[ 3].PmnPFS_BY = 0b00000101;
  R_PFS->PORT[1].PIN[ 2].PmnPFS_BY = 0b00000101;
  R_PFS->PORT[1].PIN[ 6].PmnPFS_BY = 0b00000101;
  R_PFS->PORT[1].PIN[ 7].PmnPFS_BY = 0b00000101;
  R_PFS->PORT[3].PIN[ 3].PmnPFS_BY = 0b00000101;
  cyber_speed2 = j;
  R_ICU->IELSR[PORT_IRQ1_ISEL] = 2; // PORT_IRQ1, 割り込み要求クリア
}

void put_atari_p1_joydrv()
{
  if(R_PFS->PORT[1].PIN[ 5].PmnPFS_BY&0x02) {
    if(joydrv_phase_cnt&0x01) joydrv_phase_cnt = 99;
  }
  else {
    if(!(joydrv_phase_cnt&0x01)) joydrv_phase_cnt = 99;
  }
  R_ICU->IELSR[PORT_IRQ0_ISEL] = 1; // PORT_IRQ0, 割り込み要求クリア
  switch(joydrv_phase_cnt) {
    case 0:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_portno &= 0x7f;
      else joydrv_rcv_portno |= 0x80;
      // P101 = Input, ATARI_P0_6
      if(R_PFS->PORT[1].PIN[ 1].PmnPFS_BY & 0x02) joydrv_rcv_portno &= 0xbf;
      else joydrv_rcv_portno |= 0x40;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000100;
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000101;
      break;
    case 1:
       // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_portno &= 0xdf;
      else joydrv_rcv_portno |= 0x20;
      // P101 = Input, ATARI_P0_6
      if(R_PFS->PORT[1].PIN[ 1].PmnPFS_BY & 0x02) joydrv_rcv_portno &= 0xef;
      else joydrv_rcv_portno |= 0x10;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000100;
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000101;
      break;
    case 2:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_portno &= 0xf7;
      else joydrv_rcv_portno |= 0x08;
      // P101 = Input, ATARI_P0_6
      if(R_PFS->PORT[1].PIN[ 1].PmnPFS_BY & 0x02) joydrv_rcv_portno &= 0xfb;
      else joydrv_rcv_portno |= 0x04;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000100;
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000101;
      break;
    case 3:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_portno &= 0xfd;
      else joydrv_rcv_portno |= 0x02;
      // P101 = Input, ATARI_P0_6
      if(R_PFS->PORT[1].PIN[ 1].PmnPFS_BY & 0x02) joydrv_rcv_portno &= 0xfe;
      else joydrv_rcv_portno |= 0x01;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000100;
      break;
    case 4:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][0]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][0]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][0]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][0]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 5:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor1 &= 0x7f;
      else joydrv_rcv_motor1 |= 0x80;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][0]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][0]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][0]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][0]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 6:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][1]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][1]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][1]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][1]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 7:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor1 &= 0xbf;
      else joydrv_rcv_motor1 |= 0x40;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][1]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][1]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][1]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][1]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 8:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][2]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][2]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][2]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][2]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 9:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor1 &= 0xdf;
      else joydrv_rcv_motor1 |= 0x20;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][2]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][2]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][2]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][2]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 10:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][3]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][3]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][3]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][3]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 11:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor1 &= 0xef;
      else joydrv_rcv_motor1 |= 0x10;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][3]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][3]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][3]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][3]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 12:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][4]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][4]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][4]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][4]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 13:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor1 &= 0xf7;
      else joydrv_rcv_motor1 |= 0x08;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][4]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][4]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][4]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][4]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 14:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][5]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][5]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][5]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][5]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 15:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor1 &= 0xfb;
      else joydrv_rcv_motor1 |= 0x04;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][5]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][5]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][5]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][5]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 16:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][6]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][6]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][6]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][6]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 17:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor1 &= 0xfd;
      else joydrv_rcv_motor1 |= 0x02;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][6]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][6]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][6]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][6]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 18:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][7]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][7]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][7]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][7]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 19:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor1 &= 0xfe;
      else joydrv_rcv_motor1 |= 0x01;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][7]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][7]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][7]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][7]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;

    case 20:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][8]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][8]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][8]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][8]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 21:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor2 &= 0x7f;
      else joydrv_rcv_motor2 |= 0x80;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][8]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][8]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][8]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][8]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 22:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][9]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][9]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][9]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][9]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 23:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor2 &= 0xbf;
      else joydrv_rcv_motor2 |= 0x40;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][9]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][9]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][9]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][9]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 24:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][10]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][10]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][10]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][10]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 25:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor2 &= 0xdf;
      else joydrv_rcv_motor2 |= 0x20;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][10]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][10]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][10]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][10]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 26:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][11]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][11]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][11]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][11]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 27:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor2 &= 0xef;
      else joydrv_rcv_motor2 |= 0x10;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][11]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][11]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][11]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][11]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 28:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][12]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][12]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][12]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][12]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 29:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor2 &= 0xf7;
      else joydrv_rcv_motor2 |= 0x08;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][12]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][12]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][12]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][12]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 30:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][13]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][13]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][13]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][13]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 31:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor2 &= 0xfb;
      else joydrv_rcv_motor2 |= 0x04;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][13]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][13]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][13]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][13]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 32:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][14]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][14]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][14]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][14]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 33:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor2 &= 0xfd;
      else joydrv_rcv_motor2 |= 0x02;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][14]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][14]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][14]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][14]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 34:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][15]&0x10) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][15]&0x20) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][15]&0x40) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][15]&0x80) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000100; // P101 = Output, P101= L, ATARI_P0_6
      break;
    case 35:
      // P100 = Input, ATARI_P0_7
      if(R_PFS->PORT[1].PIN[ 0].PmnPFS_BY & 0x02) joydrv_rcv_motor2 &= 0xfe;
      else joydrv_rcv_motor2 |= 0x01;
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][15]&0x01) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][15]&0x02) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][15]&0x04) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = ((joydrv_snddata[joydrv_rcv_portno][15]&0x08) ? 0b00000101 : 0b00000100);
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00000101; // P101 = Output, P101= H, ATARI_P0_6
      break;
    case 36:
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00010000; // P101 = Input, PULLUP = ON, ATARI_P0_6
      if(stick_ctrldata[joydrv_rcv_portno].motor1 != joydrv_rcv_motor1) {
        stick_ctrldata[joydrv_rcv_portno].motor1 = joydrv_rcv_motor1;
        stick_ctrldata[joydrv_rcv_portno].flg_change = true;
      }
      if(stick_ctrldata[joydrv_rcv_portno].motor2 != joydrv_rcv_motor2) {
        stick_ctrldata[joydrv_rcv_portno].motor2 = joydrv_rcv_motor2;
        stick_ctrldata[joydrv_rcv_portno].flg_change = true;
      }
      break;
    case 37:
    default:
      break;
  }
  if(++joydrv_phase_cnt>37) {
    if(joydrv_phase_cnt>99) {
      R_PFS->PORT[0].PIN[14].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 0].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 1].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[0].PIN[ 2].PmnPFS_BY = 0b00000101;
      R_PFS->PORT[1].PIN[ 1].PmnPFS_BY = 0b00010000; // P101 = Input, PULLUP = ON, ATARI_P0_6
    }
    joydrv_phase_cnt=0;
  }
  joydrv_connect_cnt=0;
//  R_ICU->IELSR[PORT_IRQ0_ISEL] = 1; // PORT_IRQ0, 割り込み要求クリア
}
