// USB controller driver ver 3.00 (UNO R4用)
// Copyright
//   メインプログラム :
//     HUYE 2020-2024
// Based on　hiduniversal.cpp

#include "joyuniversal.h"

extern unsigned int usbtask_time;
extern void clr_joydata(byte, byte);

bool JOYUniversal::bResetInitiated = false;

JOYUniversal::JOYUniversal(USB *p) :
pUsb(p),
bAddress(0),
bNbrPorts(0),
bPollEnable(false),
bHasReportId(false) {
        Initialize();
        PORT_NO = -1;

        if(pUsb)
                pUsb->RegisterDeviceClass(this);
}

void JOYUniversal::Initialize() {
        uint8_t i,j;

        for(i = 0; i < maxHidInterfaces; i++) {
                hidInterfaces[i].bmInterface = 0;
                hidInterfaces[i].bmProtocol = 0;

                for(j = 0; j < maxEpPerInterface; j++)
                        hidInterfaces[i].epIndex[j] = 0;
        }
        for(i = 0; i < totalEndpoints; i++) {
                epInfo[i].epAddr = i;
                epInfo[i].maxPktSize = 8;
//                epInfo[i].maxPktSize = (i) ? 0 : 8;
                epInfo[i].bmSndToggle = 0;
                epInfo[i].bmRcvToggle = 0;
                epInfo[i].bmNakPower = (i) ? USB_NAK_NOWAIT : USB_NAK_MAX_POWER;
                pollInterval[i] = 0;
                qNextPollTime[i] = 0;
        }

        for(i = 0; i < constBuffLen; i++)
                prevBuf[i] = 0;
        
        bNumEP = 1;
        bNumIface = 0;
        bConfNum = 0;
}

uint8_t JOYUniversal::Init(uint8_t parent, uint8_t port, bool lowspeed) {
        const static uint16_t Tbl_cnv_data[][6] PROGMEM = {
                {0x054c, 0x09cc, 5 , 1, 2, TYPE_PS4}, // PS4標準コントローラ
                {0x0ca3, 0x0024, 1 , 1, 2, TYPE_MDmini}, // MDmini標準コントローラ
                {0x0ca3, 0x0027, 1 , 1, 2, TYPE_MDmini}, // ASTRO CITY CONTROL PAD :Added by たねけん
                {0x054c, 0x0ba0, 1 , 1, 2, TYPE_PS4}, // PS4 Wireless Adapter
                {0x054c, 0x0cda, 1 , 1, 2, TYPE_PSC}, // PlayStation Classic USB Controller
                {0x054c, 0x0268, 3 , 2, 1, TYPE_PS3}, // PS3 DUALSHOCK3 Controller
                {0x0f0d, 0x0011, 1 , 1, 2, TYPE_RAP3}, // HORI Real Arcade Pro.3 SA PS3コントローラ
                {0x0f0d, 0x0022, 1 , 1, 2, TYPE_RAP3}, // HORI Real Arcade Pro.3 SA PS3コントローラ : Thanks to 佐久間亮介
                {0x0f0d, 0x0010, 1 , 1, 2, TYPE_RAP3}, // HORI FIGHTING STICK3(HORI) PS3コントローラ : Thanks to コケガエル
                {0x0f0d, 0x008b, 1 , 1, 2, TYPE_RAP3}, // HORI RAP V HAYABUSA Controller(PS3)
                {0x0f0d, 0x0040, 1 , 1, 2, TYPE_RAP3}, // HORI FIGHTING STICK mini3(HORI) PS3コントローラ : Thanks to コケガエル
                {0x0f0d, 0x008a, 1 , 1, 2, TYPE_PS4},  // HORI RAP V HAYABUSA Controller(PS4)
                {0x0f0d, 0x0066, 1 , 1, 2, TYPE_PS4},  // HORIPAD FPS+(PS4)
                {0x0f0d, 0x00ee, 1 , 1, 2, TYPE_PS4},  // HORI ワイヤードコントローラライト for PS4-102
                {0x0583, 0x2060, 1 , 1, 2, TYPE_SNES}, // iBUFFALO SNES CLASSIC USB GAMEPAD
                {0x1345, 0x1030, 1 , 1, 2, TYPE_SNES}, // RetroFreak GAME CONTROLLER
                {0x0413, 0x502b, 1 , 1, 2, TYPE_RETROFREAK}, // RetroFreak CONTROLLER ADAPTER
                {0x045e, 0x02d1, 2 , 2, 1, TYPE_XBOXONE}, // XBOX One 標準コントローラ
                {0x045e, 0x0b12, 2 , 2, 1, TYPE_XBOXONE}, // XBOX ワイヤレスコントローラ
                {0x045e, 0x028e, 0 , 1, 2, TYPE_XBOX360}, // XBOX360 標準コントローラ
                {0x1bad, 0xf503, 0 , 1, 2, TYPE_XBOX360}, // XBOX360 HORI VX
                {0x0f0d, 0x000d, 0 , 1, 2, TYPE_XBOX360}, // XBOX360 HORI FIGHTING STICK EX2
                {0x0f0d, 0x0016, 0 , 1, 2, TYPE_XBOX360}, // XBOX360 REAL ARCADE PRO.EX : Thanks to コケガエル
                {0x2dc8, 0x3100, 1 , 1, 2, TYPE_8BITDO_TYPE1_DI}, // 8BitDo DirectPad
                {0x2dc8, 0x3107, 1 , 1, 2, TYPE_8BITDO_TYPE1_DI}, // 8BitDo DirectPad
                {0x0f0d, 0x0138, 1 , 1, 2, TYPE_PCENGNmini}, // PCエンジンmini標準コントーラ
                {0x0c12, 0x0bf2, 1 , 1, 2, TYPE_NEOGEOmini}, // NEOGEOminiスーパーコンバータ
                {0x056e, 0x00e3, 1 , 1, 2, TYPE_MOUSE1}, // マウス(ELECOM M-XGL10UB)
                {0x056e, 0x011a, 1 , 1, 2, TYPE_MOUSE4}, // マウス(ELECOM M-XGL10UB)
                {0x093a, 0x2510, 1 , 1, 2, TYPE_MOUSE2}, // マウス(BUFFALO BSMOUK01)
                {0x045e, 0x0039, 1 , 1, 2, TYPE_MOUSE2}, // マウス(MS IntelliMouse Optical)
                {0x054c, 0x0ce6, 0 , 1, 2, TYPE_PS5}, // PS5標準コントローラ
                {0x0ae4, 0x0701, 1 , 1, 2, TYPE_MOUSE3}, // イーグレットⅡミニ パドルコントローラ
                {0x0ae4, 0x0702, 1 , 1, 2, TYPE_EGRETII}, // Taito EGRET II Mini Control Pad :Added by たねけん
                {0x056e, 0x200F, 0 , 1, 2, TYPE_ELECOM1}, // JC-U4013SBK(DirectInput) : Thanks to みゆ
                {0x056e, 0x2012, 4 , 1, 2, TYPE_ELECOM2}, // JC-U4013SBK(XInput) : Thanks to みゆ
                {0x05b8, 0x1004, 0 , 1, 2, TYPE_ELECOM3}, // SYSGRATION ELECOM JC-U3312 : Thanks to kani
                {0x056e, 0x200E, 0 , 1, 2, TYPE_ELECOM4}, // JC-U3912TBK(DirectInput) : Thanks to taneken
                {0x0079, 0x1830, 1 , 1, 2, TYPE_MYFLASHF300_DI}, // MAYFLASH ARCADE STICK F300(DINPUT/PS3)
                {0x0079, 0x181c, 1 , 1, 2, TYPE_MYFLASHF300_XB}, // MAYFLASH ARCADE STICK F300(XBOX)
                {0x20d6, 0xa712, 1 , 1, 2, TYPE_MYFLASHF300_DI}, // MAYFLASH ARCADE STICK F300(XINPUT/NS)
                {0x17a7, 0x0005, 1 , 1, 2, TYPE_CYBERSTICK}, // Cyber Stick(マイコンソフト)
                {0x2c22, 0x2300, 1 , 1, 2, TYPE_PS4}, // QANBA OBSIDIAN(PS4)
                {0x2c22, 0x2302, 1 , 1, 2, TYPE_RAP3}, // QANBA OBSIDIAN(PS3)
                {0x2c22, 0x2303, 0 , 1, 2, TYPE_XBOX360}, // QANBA OBSIDIAN(PC)
                {0x0603, 0x6875, 0 , 1, 2, TYPE_SANWA1}, // JY-P35U(DirectInput) : Thanks to kehu
                {0x33dd, 0x0013, 0 , 1, 2, TYPE_ZUIKI1}, // X68000Zコントローラ
                {0xffff, 0     , 1 , 1, 2, TYPE_DIRECTINPUT} // データ終端
        };

        uint8_t buf[constBufSize];
//        uint8_t buf2[24];
        USB_DEVICE_DESCRIPTOR * udd = reinterpret_cast<USB_DEVICE_DESCRIPTOR*>(buf);
        HubDescriptor* hd = reinterpret_cast<HubDescriptor*>(buf);
        USB_CONFIGURATION_DESCRIPTOR * ucd = reinterpret_cast<USB_CONFIGURATION_DESCRIPTOR*>(buf);
        uint8_t rcode;
        int cnv_pointer;
        UsbDevice *p = NULL;
        EpInfo *oldep_ptr = NULL;
        uint8_t len = 0;
        uint16_t cd_len = 0;
        uint8_t writeBuf[256];

        uint8_t num_of_conf; // number of configurations
        uint8_t i,j;

//        TIMSK0 |= B00000001;  // Timer0割り込み許可

        AddressPool &addrPool = pUsb->GetAddressPool();

        if(bAddress)
                return USB_ERROR_CLASS_INSTANCE_ALREADY_IN_USE;

        // Get pointer to pseudo device with address 0 assigned
        p = addrPool.GetUsbDevicePtr(0);

        if(!p)
                return USB_ERROR_ADDRESS_NOT_FOUND_IN_POOL;

        if(!p->epinfo) 
                return USB_ERROR_EPINFO_IS_NULL;

        // Save old pointer to EP_RECORD of address 0
        oldep_ptr = p->epinfo;

        // Temporary assign new pointer to epInfo to p->epinfo in order to avoid toggle inconsistence
        p->epinfo = epInfo;

        p->lowspeed = lowspeed;
        // Get device descriptor
        rcode = pUsb->getDevDescr(0, 0, 8, (uint8_t*)buf);

        p->lowspeed = ((udd->bcdUSB < 0x0200) ? lowspeed : false);

        if(!rcode)
                len = (buf[0] > constBufSize) ? constBufSize : buf[0];

        if(rcode) {
                // Restore p->epinfo
                p->epinfo = oldep_ptr;

                goto FailGetDevDescr;
        }

        // Allocate new address according to device class
        bAddress = addrPool.AllocAddress(parent, (udd->bDeviceClass == 0x09) ? true : false, port);
//Serial.print(F("bAddress:"));
//Serial.println(bAddress, HEX);

        if(!bAddress)
                return USB_ERROR_OUT_OF_ADDRESS_SPACE_IN_POOL;

        // Extract Max Packet Size from the device descriptor
        epInfo[0].maxPktSize = udd->bMaxPacketSize0;

        // Assign new address to the device
//        rcode = pUsb->setAddr(0, 0, bAddress);
        rcode = pUsb->ctrlReq(0, 0, bmREQ_SET, USB_REQUEST_SET_ADDRESS, bAddress, 0x00, 0x0000, 0x0000, 0x0000, NULL, NULL);
        delay(300);

        // Restore p->epinfo
        p->epinfo = oldep_ptr;

        if(rcode) {
                addrPool.FreeAddress(bAddress);
                bAddress = 0;
                return rcode;
        }

        if(udd->bDeviceClass != 0x09) {
                p = addrPool.GetUsbDevicePtr(bAddress);
                if(!p)
                        return USB_ERROR_ADDRESS_NOT_FOUND_IN_POOL;
                p->lowspeed = lowspeed;
        }

        if(len)
                rcode = pUsb->getDevDescr(bAddress, 0, len, (uint8_t*)buf);

        if(rcode)
                goto FailGetDevDescr;

        if(udd->bDeviceClass != 0x09) {
                VID = udd->idVendor; // Can be used by classes that inherits this class to check the VID and PID of the connected device
                PID = udd->idProduct;
//Serial.print(F("VID:"));
//Serial.println(VID, HEX);
//Serial.print(F("PID:"));
//Serial.println(PID, HEX);

                cnv_pointer=0;
                while(pgm_read_word_near(&Tbl_cnv_data[cnv_pointer][0]) != 0xffff) {
                        if(pgm_read_word_near(&Tbl_cnv_data[cnv_pointer][0]) == VID && pgm_read_word_near(&Tbl_cnv_data[cnv_pointer][1]) == PID) break;
                        cnv_pointer++;
                }
                DTYPE  = pgm_read_word_near(&Tbl_cnv_data[cnv_pointer][2]);
                JOYTYPE = pgm_read_word_near(&Tbl_cnv_data[cnv_pointer][5]);
//Serial.print(F("DTYPE:"));
//Serial.println(DTYPE, HEX);
//Serial.print(F("JOYTYPE:"));
//Serial.println(JOYTYPE, HEX);

                epInterruptInIndex  = pgm_read_word_near(&Tbl_cnv_data[cnv_pointer][3]);
                epInterruptOutIndex = pgm_read_word_near(&Tbl_cnv_data[cnv_pointer][4]);
//Serial.print(F("bDeviceClass:"));
//Serial.println(udd->bDeviceClass, HEX);

                num_of_conf = udd->bNumConfigurations;
//Serial.print(F("num_of_conf:"));
//Serial.println(num_of_conf, HEX);

                // Assign epInfo to epinfo pointer
                rcode = pUsb->setEpInfoEntry(bAddress, 1, epInfo);
                if(rcode)
                        goto FailSetDevTblEntry;

                for(i = 0; i < num_of_conf; i++) {
                        ConfigDescParser<0, 0, 0, 0> confDescrParser(this);
                        rcode = pUsb->getConfDescr(bAddress, 0, i, &confDescrParser);
                        if(rcode)
                                goto FailGetConfDescr;
                        if(bNumEP >= totalEndpoints)
                        break;
 
                } // for
//Serial.print(F("bNumEP:"));
//Serial.println(bNumEP, HEX);

                if(bNumEP < 2)
                        goto FailUnknownDevice;

                // Assign epInfo to epinfo pointer
                rcode = pUsb->setEpInfoEntry(bAddress, bNumEP, epInfo);
                if(rcode)
                        goto FailSetDevTblEntry;

                delay(200);

                rcode = pUsb->setConf(bAddress, 0, bConfNum);
                if(rcode)
                        goto FailSetConfDescr;

                delay(200);

//Serial.print(F("bConfNume:"));
//Serial.println(bConfNum, HEX);
//Serial.print(F("bNumIface:"));
//Serial.println(bNumIface, HEX);
                for(i = 0; i < bNumIface; i++) {
//Serial.print(F("i:"));
//Serial.println(i, HEX);
//Serial.print(F("hidInterfaces[i].bmInterface:"));
//Serial.println(hidInterfaces[i].bmInterface, HEX);
//Serial.print(F("hidInterfaces[i].bmProtocol:"));
//Serial.println(hidInterfaces[i].bmProtocol, HEX);
//Serial.print(F("hidInterfaces[i].epIndex[epInterruptInIndex]:"));
//Serial.println(hidInterfaces[i].epIndex[epInterruptInIndex], HEX);
//Serial.print(F("epInfo[hidInterfaces[i].epIndex[epInterruptInIndex]].epAddr:"));
//Serial.println(epInfo[hidInterfaces[i].epIndex[epInterruptInIndex]].epAddr, HEX);
//Serial.print(F("hidInterfaces[i].epIndex[epInterruptOutIndex]:"));
//Serial.println(hidInterfaces[i].epIndex[epInterruptOutIndex], HEX);
//Serial.print(F("epInfo[hidInterfaces[i].epIndex[epInterruptOutIndex]].epAddr:"));
//Serial.println(epInfo[hidInterfaces[i].epIndex[epInterruptOutIndex]].epAddr, HEX);

                        if(hidInterfaces[i].epIndex[epInterruptInIndex] == 0)
                                continue;
                        switch(DTYPE) {
                                case 1:
                                case 5:
                                        rcode = pUsb->ctrlReq(bAddress, 0, bmREQ_HID_OUT, HID_REQUEST_SET_IDLE, 0, 0, hidInterfaces[i].bmInterface, 0x0000, 0x0000, NULL, NULL);
                                        if(rcode && rcode != hrSTALL)
                                                goto FailSetIdle;
                                        if(DTYPE == 1) {
                                                rcode = pUsb->ctrlReq(bAddress, 0, bmREQ_HID_REPORT, USB_REQUEST_GET_DESCRIPTOR, 0, HID_DESCRIPTOR_REPORT, hidInterfaces[i].bmInterface, 256, 256, writeBuf, NULL);
                                                if(rcode && rcode != hrSTALL)
                                                        goto FailSetIdle;
                                        }
                                        break;
                                case 2:
                                        cmdCounter = 0; // Reset the counter used when sending out the commands
                                        writeBuf[0] = 0x05;
                                        writeBuf[1] = 0x20;
                                        writeBuf[2] = cmdCounter++;
                                        writeBuf[3] = 0x01;
                                        writeBuf[4] = 0x00;
                                        SndRpt(5, writeBuf);
                                        break;
                                case 3:
                                        writeBuf[0] = 0x42;
                                        writeBuf[1] = 0x0c;
                                        writeBuf[2] = 0x00;
                                        writeBuf[3] = 0x00;
                                        rcode = pUsb->ctrlReq(bAddress, 0, bmREQ_HID_OUT, HID_REQUEST_SET_REPORT, 0xF4, 0x03, 0x00, 4, 4, writeBuf, NULL);
                                        if(rcode && rcode != hrSTALL)
                                                goto FailSetIdle;

                                        for(j = 0; j < PS3_REPORT_BUFFER_SIZE; j++)
                                                writeBuf[j] = pgm_read_byte(&ps3_report_buffer[j]);

                                        rcode = pUsb->ctrlReq(bAddress, 0, bmREQ_HID_OUT, HID_REQUEST_SET_REPORT, 0x01, 0x02, 0x00, PS3_REPORT_BUFFER_SIZE, PS3_REPORT_BUFFER_SIZE, writeBuf, NULL);
                                        if(rcode && rcode != hrSTALL)
                                                goto FailSetIdle;
                                        break;
                                case 4:
                                        rcode = pUsb->ctrlReq(bAddress, 0, USB_SETUP_DEVICE_TO_HOST|USB_SETUP_TYPE_VENDOR|USB_SETUP_RECIPIENT_INTERFACE, USB_REQUEST_CLEAR_FEATURE, 0, 1, hidInterfaces[i].bmInterface, 20, 0, NULL, NULL);
                                        if(rcode && rcode != hrSTALL)
                                                goto FailSetIdle;
                                        rcode = pUsb->ctrlReq(bAddress, 0, USB_SETUP_DEVICE_TO_HOST|USB_SETUP_TYPE_VENDOR|USB_SETUP_RECIPIENT_INTERFACE, USB_REQUEST_CLEAR_FEATURE, 0, 0, hidInterfaces[i].bmInterface,  8, 0, NULL, NULL);
                                        if(rcode && rcode != hrSTALL)
                                                goto FailSetIdle;
                                        rcode = pUsb->ctrlReq(bAddress, 0, USB_SETUP_DEVICE_TO_HOST|USB_SETUP_TYPE_VENDOR|USB_SETUP_RECIPIENT_DEVICE, USB_REQUEST_CLEAR_FEATURE, 0, 0, hidInterfaces[i].bmInterface,  4, 0, NULL, NULL);
                                        if(rcode && rcode != hrSTALL)
                                                goto FailSetIdle;
                                        break;

                        }
                }
                PORT_NO = (bAddress>=9 ? bAddress-9 : 0);
                clr_joydata(PORT_NO, 0);
        }
        else {
//Serial.println(F("DEVICE HUB"));
                // Assign epInfo to epinfo pointer
                rcode = pUsb->setEpInfoEntry(bAddress, 2, epInfo);
                if(rcode)
                        goto FailSetDevTblEntry;

//Serial.println(F("HUB PHASE1"));
                rcode = GetHubDescriptor(0, 8, buf);
                if(rcode)
                        goto FailGetHubDescr;

//Serial.println(F("HUB PHASE2"));
                // Save number of ports for future use
                bNbrPorts = hd->bNbrPorts;
//Serial.print(F("bNbrPorts:"));
//Serial.println(bNbrPorts, HEX);
                // Read configuration Descriptor in Order To Obtain Proper Configuration Value
                rcode = pUsb->getConfDescr(bAddress, 0, 8, 0, buf);
                if(!rcode) {
                        cd_len = ucd->wTotalLength;
//Serial.print(F("cd_len:"));
//Serial.println(cd_len, HEX);
                        rcode = pUsb->getConfDescr(bAddress, 0, cd_len, 0, buf);
                }
                if(rcode)
                        goto FailGetConfDescr;

//Serial.println(F("HUB PHASE3"));
                // The following code is of no practical use in real life applications.
                // It only intended for the usb protocol sniffer to properly parse hub-class requests.
//                rcode = pUsb->getConfDescr(bAddress, 0, buf[0], 0, buf2);
//                if(rcode)
//                        goto FailGetConfDescr;

//Serial.println(F("HUB PHASE4"));
                // Set Configuration Value
//Serial.print(F("buf[5]:"));
//Serial.println(buf[5], HEX);
                rcode = pUsb->setConf(bAddress, 0, buf[5]);
                if(rcode)
                        goto FailSetConfDescr;

//Serial.println(F("HUB PHASE5"));
                // Power on all ports
                for(j = 1; j <= bNbrPorts; j++)
                        SetPortFeature(HUB_FEATURE_PORT_POWER, j, 0); //HubPortPowerOn(j);

                pUsb->SetHubPreMask();
        }
        bPollEnable = true;
        //TIMSK0 &= 0b11111110;  // Timer0割り込み禁止
        return 0;

FailUnknownDevice:
        rcode = USB_DEV_CONFIG_ERROR_DEVICE_NOT_SUPPORTED;
#ifdef DEBUG_USB_HOST
        goto Fail;
#endif

FailGetDevDescr:
#ifdef DEBUG_USB_HOST
        NotifyFailGetDevDescr();
        goto Fail;
#endif

FailSetDevTblEntry:
#ifdef DEBUG_USB_HOST
        NotifyFailSetDevTblEntry();
        goto Fail;
#endif

FailGetHubDescr:
#ifdef DEBUG_USB_HOST
        NotifyFailGetConfDescr();
        goto Fail;
#endif

FailGetConfDescr:
#ifdef DEBUG_USB_HOST
        NotifyFailGetConfDescr();
        goto Fail;
#endif

FailSetConfDescr:
#ifdef DEBUG_USB_HOST
        NotifyFailSetConfDescr();
        goto Fail;
#endif

FailSetIdle:

#ifdef DEBUG_USB_HOST
Fail:
        NotifyFail(rcode);
#endif
        Release();
        //TIMSK0 &= 0b11111110;  // Timer0割り込み禁止
        return rcode;
}


JOYUniversal::HIDInterface* JOYUniversal::FindInterface(uint8_t iface, uint8_t alt, uint8_t proto) {
        for(uint8_t i = 0; i < bNumIface; i++) {
                if(hidInterfaces[i].bmInterface == iface && hidInterfaces[i].bmAltSet == alt
                && hidInterfaces[i].bmProtocol == proto)
                        return hidInterfaces + i;
        }
        return NULL;
}

void JOYUniversal::EndpointXtract(uint8_t conf, uint8_t iface, uint8_t alt, uint8_t proto, const USB_ENDPOINT_DESCRIPTOR *pep) {
        // If the first configuration satisfies, the others are not concidered.
//Serial.print(F("conf:"));
//Serial.println(conf, HEX);
//Serial.print(F("bConfNum:"));
//Serial.println(bConfNum, HEX);

        if(bNumIface >= maxHidInterfaces)
                return;

        bConfNum = conf;

        if(alt)
                return;

        uint8_t index;

        if((pep->bmAttributes & bmUSB_TRANSFER_TYPE) == USB_TRANSFER_TYPE_INTERRUPT) {
                index = (((pep->bEndpointAddress & 0x80) == 0x80) ? epInterruptInIndex : epInterruptOutIndex);

                HIDInterface *piface = FindInterface(iface, alt, proto);
//Serial.print(F("iface:"));
//Serial.println(iface, HEX);
//Serial.print(F("proto:"));
//Serial.println(proto, HEX);
//Serial.print(F("bNumIface:"));
//Serial.println(bNumIface, HEX);

                // Fill in interface structure in case of new interface
                if(!piface) {
                        piface = hidInterfaces + bNumIface;
                        piface->bmInterface = iface;
                        piface->bmAltSet = alt;
                        piface->bmProtocol = proto;
                        bNumIface++;
                }

                // Fill in the endpoint info structure
                epInfo[bNumEP].epAddr = (pep->bEndpointAddress & 0x0F);
                epInfo[bNumEP].epAttribs = USB_TRANSFER_TYPE_INTERRUPT;
                epInfo[bNumEP].maxPktSize = (uint8_t)pep->wMaxPacketSize;
                epInfo[bNumEP].bmSndToggle = 0;
                epInfo[bNumEP].bmRcvToggle = 0;
                epInfo[bNumEP].bmNakPower = USB_NAK_NOWAIT;
                pollInterval[bNumEP] = pep->bInterval; // Set the polling interval as the largest polling interval obtained from endpoints
                if(JOYTYPE == TYPE_MOUSE3) pollInterval[bNumEP] = 60;
                else if(JOYTYPE == TYPE_MOUSE1 || JOYTYPE == TYPE_MOUSE2 || JOYTYPE == TYPE_MOUSE4) pollInterval[bNumEP] = 10;
                if(JOYTYPE == TYPE_CYBERSTICK) pollInterval[bNumEP] = 4;

//Serial.print(F("index:"));
//Serial.println(index, HEX);
//Serial.print(F("bNumEP:"));
//Serial.println(bNumEP, HEX);
//Serial.print(F("epInfo[bNumEP].epAddr:"));
//Serial.println(epInfo[bNumEP].epAddr, HEX);
//Serial.print(F("pollInterval[bNumEP]:"));
//Serial.println(pollInterval[bNumEP], HEX);

                // Fill in the endpoint index list
                piface->epIndex[index] = bNumEP; //(pep->bEndpointAddress & 0x0F);

                bNumEP++;
        }
        //PrintEndpointDescriptor(pep);
}

uint8_t JOYUniversal::Release() {
        pUsb->GetAddressPool().FreeAddress(bAddress);

        if(bAddress == 0x41)
                pUsb->SetHubPreMask();
        Initialize();
        if(PORT_NO >= 0) clr_joydata(PORT_NO, 0);

        PORT_NO = -1;
        bAddress = 0;
        bNbrPorts = 0;
        for(uint8_t i = 0; i < totalEndpoints; i++) qNextPollTime[i] = 0;
        bPollEnable = false;
        return 0;
}

uint8_t JOYUniversal::Poll() {
        uint8_t rcode = 0;
        uint8_t buf[constBuffLen];
        uint16_t read;

        if(!bPollEnable)
                return 0;

        if(bAddress != 0x41) {
                for(uint8_t i = 0; i < bNumIface; i++) {
                        uint8_t index = hidInterfaces[i].epIndex[epInterruptInIndex];
                        if(!index)
                                continue;

                        qNextPollTime[index]-=usbtask_time; // 1ms
                        if(qNextPollTime[index] <= 0) {
                                qNextPollTime[index] = (int16_t)pollInterval[index];
//                        if((int32_t)((uint32_t)millis() - qNextPollTime[index]) >= 0L) {
//                                qNextPollTime[index] = (uint32_t)millis() + (uint32_t)pollInterval[index];

                                read = (uint16_t)epInfo[index].maxPktSize;
                                for(uint8_t j = 0; j < constBuffLen; j++)
                                        buf[j] = 0;

                                rcode = pUsb->inTransfer(bAddress, epInfo[index].epAddr, &read, buf);
                                if(rcode == 0) {
                                        if(read > constBuffLen)
                                                read = constBuffLen;

                                        bool identical = true;
                                        for(uint8_t j = 0; j < read; j++) {
                                                if(buf[j] != prevBuf[j]) {
                                                        identical = false;
                                                        prevBuf[j] = buf[j];
                                                }
                                        }
                                        if(identical)
                                                continue;
                                        ParseHIDData(this, bHasReportId, (uint8_t)read, buf);
                                }
                                else if(rcode != hrNAK) break;
                                else {
                                  if(JOYTYPE == TYPE_MOUSE1 || JOYTYPE == TYPE_MOUSE2 || JOYTYPE == TYPE_MOUSE4) clr_joydata(PORT_NO, 1);
                                  rcode = 0;
                                }
                        }
                }
        }
        else {
                  qNextPollTime[0]-=usbtask_time; // 1ms
                  if(qNextPollTime[0] <= 0) {
//                if(((int32_t)((uint32_t)millis() - qNextPollTime[0]) >= 0L)) {
                        rcode = CheckHubStatus();
                        qNextPollTime[0] = 100;
//                        qNextPollTime[0] = (uint32_t)millis() + 100;
                  }
        }
        return rcode;
}

uint8_t JOYUniversal::CheckHubStatus() {
        uint8_t rcode;
        uint8_t buf[8];
        uint16_t read = 1;

        rcode = pUsb->inTransfer(bAddress, 1, &read, buf);
//Serial.print(F("HUB inTransfer rcode:"));
//Serial.println(rcode, HEX);

        if(rcode)
                return rcode;

//Serial.println(F("HUB POOL"));
        for(uint8_t port = 1, mask = 0x02; port < 8; mask <<= 1, port++) {
                if(buf[0] & mask) {
                        HubEvent evt;
                        evt.bmEvent = 0;

                        rcode = GetPortStatus(port, 4, evt.evtBuff);

                        if(rcode)
                                continue;

                        rcode = PortStatusChange(port, evt);

                        if(rcode == HUB_ERROR_PORT_HAS_BEEN_RESET)
                                return 0;

                        if(rcode)
                                return rcode;
                }
        }

        for(uint8_t port = 1; port <= bNbrPorts; port++) {
                HubEvent evt;
                evt.bmEvent = 0;

                rcode = GetPortStatus(port, 4, evt.evtBuff);

                if(rcode)
                        continue;

                if((evt.bmStatus & bmHUB_PORT_STATE_CHECK_DISABLED) != bmHUB_PORT_STATE_DISABLED)
                        continue;

                // Emulate connection event for the port
                evt.bmChange |= bmHUB_PORT_STATUS_C_PORT_CONNECTION;

                rcode = PortStatusChange(port, evt);

                if(rcode == HUB_ERROR_PORT_HAS_BEEN_RESET)
                        return 0;

                if(rcode)
                        return rcode;
        }
        return 0;
}

void JOYUniversal::ResetHubPort(uint8_t port) {
        HubEvent evt;
        evt.bmEvent = 0;
        uint8_t rcode;

        ClearPortFeature(HUB_FEATURE_C_PORT_ENABLE, port, 0);
        ClearPortFeature(HUB_FEATURE_C_PORT_CONNECTION, port, 0);
        SetPortFeature(HUB_FEATURE_PORT_RESET, port, 0);


        for(byte i = 0; i < 3; i++) {
                rcode = GetPortStatus(port, 4, evt.evtBuff);
                if(rcode) break; // Some kind of error, bail.
                if(evt.bmEvent == bmHUB_PORT_EVENT_RESET_COMPLETE || evt.bmEvent == bmHUB_PORT_EVENT_LS_RESET_COMPLETE) {
                        break;
                }
                delay(100);
        }
        ClearPortFeature(HUB_FEATURE_C_PORT_RESET, port, 0);
        ClearPortFeature(HUB_FEATURE_C_PORT_CONNECTION, port, 0);
        delay(20);
}

uint8_t JOYUniversal::PortStatusChange(uint8_t port, HubEvent &evt) {
        switch(evt.bmEvent) {
                        // Device connected event
                case bmHUB_PORT_EVENT_CONNECT:
                case bmHUB_PORT_EVENT_LS_CONNECT:
                        if(bResetInitiated)
                                return 0;

                        ClearPortFeature(HUB_FEATURE_C_PORT_ENABLE, port, 0);
                        ClearPortFeature(HUB_FEATURE_C_PORT_CONNECTION, port, 0);
                        SetPortFeature(HUB_FEATURE_PORT_RESET, port, 0);
                        bResetInitiated = true;
                        return HUB_ERROR_PORT_HAS_BEEN_RESET;

                        // Device disconnected event
                case bmHUB_PORT_EVENT_DISCONNECT:
                        ClearPortFeature(HUB_FEATURE_C_PORT_ENABLE, port, 0);
                        ClearPortFeature(HUB_FEATURE_C_PORT_CONNECTION, port, 0);
                        bResetInitiated = false;

                        UsbDeviceAddress a;
                        a.devAddress = 0;
                        a.bmHub = 0;
                        a.bmParent = bAddress;
                        a.bmAddress = port;
                        pUsb->ReleaseDevice(a.devAddress);
                        return 0;

                        // Reset complete event
                case bmHUB_PORT_EVENT_RESET_COMPLETE:
                case bmHUB_PORT_EVENT_LS_RESET_COMPLETE:
                        ClearPortFeature(HUB_FEATURE_C_PORT_RESET, port, 0);
                        ClearPortFeature(HUB_FEATURE_C_PORT_CONNECTION, port, 0);

                        delay(20);

                        a.devAddress = bAddress;

                        pUsb->Configuring(a.bmAddress, port, (evt.bmStatus & bmHUB_PORT_STATUS_PORT_LOW_SPEED));
                        bResetInitiated = false;
                        break;

        } // switch (evt.bmEvent)
        return 0;
}

// Send a report to interrupt out endpoint. This is NOT SetReport() request!
uint8_t JOYUniversal::SndRpt(uint16_t nbytes, uint8_t *dataptr) {
        for(uint8_t i = 0; i < bNumIface; i++) {
                if(hidInterfaces[i].epIndex[epInterruptOutIndex] != 0)
                        return pUsb->outTransfer(bAddress, epInfo[hidInterfaces[i].epIndex[epInterruptOutIndex]].epAddr, nbytes, dataptr);
        }
        return 0;
}

uint8_t JOYUniversal::SndRpt_ps3(uint16_t nbytes, uint8_t *dataptr) {
        return pUsb->ctrlReq(bAddress, 0, bmREQ_HID_OUT, HID_REQUEST_SET_REPORT, 0x01, 0x02, 0x00, nbytes, nbytes, dataptr, NULL);
}
